import platform
from .ldap_enumerator import *

if platform.system() == 'Windows':
	from .windows.session_enumerator import *
	from .windows.share_enumerator import *
	
else:
	from .linux.session_enumerator import *
	from .linux.share_enumerator import *

__all__ = ['']


from . import Basemodel
import datetime
from sqlalchemy import Column, Integer, String, DateTime

class LocalGroup(Basemodel):
	__tablename__ = 'localgroup'
	
	id = Column(Integer, primary_key=True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	machine_id = Column(Integer)
	ip = Column(String, index=True)
	rdns = Column(String, index=True)
	hostname = Column(String, index=True)
	sid = Column(String, index=True)
	sidusage = Column(String, index=True)
	domain = Column(String, index=True)
	username = Column(String, index=True)
	groupname = Column(String, index=True)


from . import Basemodel
import datetime
from sqlalchemy import Column, Integer, String, Boolean

class HashEntry(Basemodel):
	__tablename__ = 'hashes'
	
	id = Column(Integer, primary_key=True)
	nt_hash = Column(String, index = True)
	lm_hash = Column(String, index = True)
	plaintext = Column(String, unique = True, nullable= False)
	pw_length = Column(Integer, index = True)
	pw_lower = Column(Boolean, index = True)
	pw_upper = Column(Boolean, index = True)
	pw_digit = Column(Boolean, index = True)
	pw_special = Column(Boolean, index = True)
	
	

	def isspecial(s):
		if s.isupper() or s.islower() or s in '0123456789':
			return False
		return True

	def __init__(self, plaintext, nt_hash = None, lm_hash = None):
		self.nt_hash = nt_hash
		self.lm_hash = lm_hash

		self.plaintext = None
		self.pw_length = None
		self.pw_lower = None
		self.pw_upper = None
		self.pw_upper = None
		self.pw_digit = None
		self.pw_special = None

		self.set_stats(plaintext)
		

	def set_stats(self, plaintext):
		decoded = None
		if plaintext.startswith('$HEX['):
			for enc in ['latin-1', 'utf-16-le', 'utf8']:
				try:
					decoded = bytes.fromhex(plaintext[5:-1]).decode(enc)
					break
				except Exception as e:
					continue
		else:
			decoded = plaintext

		if decoded is None:
			raise Exception('Failed to decode: %s' % plaintext)

		self.plaintext = decoded
		self.pw_length = len(self.plaintext)
		self.pw_lower = any(c.islower() for c in self.plaintext)
		self.pw_upper = any(c.isupper() for c in self.plaintext)
		self.pw_upper = any(c.isupper() for c in self.plaintext)
		self.pw_digit = any(c in '0123456789' for c in self.plaintext)
		self.pw_special = any(HashEntry.isspecial(c) for c in self.plaintext)
	
	@staticmethod
	def from_potfile(potfile):
		with open(potfile, 'r') as f:
			for line in f:
				line = line.strip()
				
				some_hash, plaintext = line.split(':', 1)				
				if len(some_hash) == 32:
					yield HashEntry(plaintext, nt_hash = some_hash)
				elif len(some_hash) == 16:
					yield HashEntry(plaintext, lm_hash = some_hash)
				else:
					continue
			


from . import Basemodel, lf
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, ForeignKey

class JackDawMachineConstrainedDelegation(Basemodel):
	__tablename__ = 'constrainedmachine'

	id = Column(Integer, primary_key=True)
	machine_id = Column(Integer, ForeignKey('machines.id'))
	machine = relationship("JackDawADMachine", back_populates="allowedtodelegateto", lazy = True)
	spn = Column(String, index=True)
	targetaccount = Column(String, index=True)


class JackDawUserConstrainedDelegation(Basemodel):
	__tablename__ = 'constraineduser'

	id = Column(Integer, primary_key=True)
	user_id = Column(Integer, ForeignKey('users.id'))
	user = relationship("JackDawADUser", back_populates="allowedtodelegateto", lazy = True)
	spn = Column(String, index=True)
	targetaccount = Column(String, index=True)


from flask import current_app
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine

def get_user_uac_problems(domainid):
	db = current_app.db
	pw_notreq_users = []
	for uid, username in db.session.query(JackDawADUser).filter_by(ad_id = domainid).filter(JackDawADUser.UAC_PASSWD_NOTREQD == True).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName).all():
		pw_notreq_users.append([uid, username])

	plaintext_pw_users = []
	for uid, username in db.session.query(JackDawADUser).filter_by(ad_id = domainid).filter(JackDawADUser.UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED == True).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName).all():
		plaintext_pw_users.append([uid, username])

	pw_notexp = []
	for uid, username in db.session.query(JackDawADUser).filter_by(ad_id = domainid).filter(JackDawADUser.UAC_DONT_EXPIRE_PASSWD == True).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName).all():
		pw_notexp.append([uid, username])

	des_only = []
	for uid, username in db.session.query(JackDawADUser).filter_by(ad_id = domainid).filter(JackDawADUser.UAC_USE_DES_KEY_ONLY == True).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName).all():
		des_only.append([uid, username])  

	asrep = []
	for uid, username in db.session.query(JackDawADUser).filter_by(ad_id = domainid).filter(JackDawADUser.UAC_DONT_REQUIRE_PREAUTH == True).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName).all():
		asrep.append([uid, username])
	
	return {
		'pw_notreq_users':  pw_notreq_users,
		'plaintext_pw_users':  plaintext_pw_users,
		'pw_notexp':  pw_notexp,
		'des_only':  des_only,
		'asrep':  asrep,
	}

def get_outdated_os(domainid):
	#TODO: implement filtering!!!
	db = current_app.db
	outdated_hosts = {} #version - samaccountname,id
	versions = {}
	for x in db.session.query(JackDawADMachine.operatingSystemVersion).filter_by(ad_id = domainid).group_by(JackDawADMachine.operatingSystemVersion).distinct():
		versions[x[0]] = 1

	#TODO: do filtering here!
	filtered_versions = versions
	for version in filtered_versions:
		#version = None if version == '' else version
		for mid, name in db.session.query(JackDawADMachine.operatingSystemVersion).filter_by(ad_id = domainid).filter(JackDawADMachine.operatingSystemVersion == version).with_entities(JackDawADMachine.id, JackDawADMachine.sAMAccountName).all():
			if version not in outdated_hosts:
				outdated_hosts[version] = []
			outdated_hosts[version].append([mid, name])
	
	return outdated_hosts



from . import Basemodel, lf
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey

class JackDawADInfo(Basemodel):
	__tablename__ = 'ads'
	
	id = Column(Integer, primary_key=True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	auditingPolicy = Column(String)
	creationTime = Column(DateTime)
	dc = Column(String)
	distinguishedName = Column(String)
	forceLogoff = Column(Integer)
	instanceType = Column(Integer)
	lockoutDuration = Column(Integer)
	lockOutObservationWindow = Column(Integer)
	lockoutThreshold = Column(Integer)
	masteredBy = Column(String)
	maxPwdAge = Column(Integer)
	minPwdAge = Column(Integer)
	minPwdLength = Column(Integer)
	name = Column(String)
	nextRid = Column(Integer)
	nTSecurityDescriptor = Column(String)
	objectCategory = Column(String)
	objectClass = Column(String)
	objectGUID = Column(String)
	objectSid = Column(String)
	pwdHistoryLength = Column(Integer)
	pwdProperties = Column(Integer)
	serverState = Column(Integer)
	systemFlags = Column(Integer)
	uASCompat = Column(Integer)
	uSNChanged = Column(Integer)
	uSNCreated = Column(Integer)
	whenChanged = Column(DateTime)
	whenCreated = Column(DateTime)

	users = relationship("JackDawADUser", back_populates="ad", lazy = True)
	computers = relationship("JackDawADMachine", back_populates="ad", lazy = True)
	groups = relationship("JackDawADGroup", back_populates="ad", lazy = True)
	group_lookups = relationship("JackDawTokenGroup", back_populates="ad", lazy = True)
	spnservices = relationship("JackDawSPNService", back_populates="ad", lazy = True)
	objectacls = relationship("JackDawADDACL", back_populates="ad", lazy = True)
	customrelations = relationship("JackDawCustomRelations", back_populates="ad", lazy = True)
	ou = relationship("JackDawADOU", back_populates="ad", lazy = True)
	
	def to_dict(self):
		return {
			'id' : self.id ,
			'fetched_at' : self.fetched_at ,
			'creationTime' : self.creationTime ,
			'distinguishedName' : self.distinguishedName ,
			'forceLogoff' : self.forceLogoff ,
			'lockoutDuration' : self.lockoutDuration ,
			'lockOutObservationWindow' : self.lockOutObservationWindow ,
			'lockoutThreshold' : self.lockoutThreshold ,
			'masteredBy' : self.masteredBy ,
			'maxPwdAge' : self.maxPwdAge ,
			'minPwdAge' : self.minPwdAge ,
			'minPwdLength' : self.minPwdLength ,
			'name' : self.name ,
			'pwdHistoryLength' : self.pwdHistoryLength ,
			'pwdProperties' : self.pwdProperties ,
			'whenChanged' : self.whenChanged ,
			'whenCreated' : self.whenCreated ,
		}

	@staticmethod
	def from_dict(d):
		adinfo = JackDawADInfo()
		adinfo.auditingPolicy = lf(d.get('auditingPolicy'))
		adinfo.creationTime = lf(d.get('creationTime'))
		adinfo.dc = lf(d.get('dc'))
		adinfo.distinguishedName = lf(d.get('distinguishedName'))
		adinfo.forceLogoff = lf(d.get('forceLogoff'))
		adinfo.instanceType = lf(d.get('instanceType'))
		adinfo.lockoutDuration = lf(d.get('lockoutDuration'))
		adinfo.lockOutObservationWindow = lf(d.get('lockOutObservationWindow'))
		adinfo.lockoutThreshold = lf(d.get('lockoutThreshold'))
		adinfo.masteredBy = lf(d.get('masteredBy'))
		adinfo.maxPwdAge = lf(d.get('maxPwdAge'))
		adinfo.minPwdAge = lf(d.get('minPwdAge'))
		adinfo.minPwdLength = lf(d.get('minPwdLength'))
		adinfo.name = lf(d.get('name'))
		adinfo.nextRid = lf(d.get('nextRid'))
		adinfo.nTSecurityDescriptor = lf(d.get('nTSecurityDescriptor'))
		adinfo.objectCategory = lf(d.get('objectCategory'))
		adinfo.objectClass = lf(d.get('objectClass'))
		adinfo.objectGUID = lf(d.get('objectGUID'))
		adinfo.objectSid = lf(d.get('objectSid'))
		adinfo.pwdHistoryLength = lf(d.get('pwdHistoryLength'))
		adinfo.pwdProperties = lf(d.get('pwdProperties'))
		adinfo.serverState = lf(d.get('serverState'))
		adinfo.systemFlags = lf(d.get('systemFlags'))
		adinfo.uASCompat = lf(d.get('uASCompat'))
		adinfo.uSNChanged = lf(d.get('uSNChanged'))
		adinfo.uSNCreated = lf(d.get('uSNCreated'))
		adinfo.whenChanged = lf(d.get('whenChanged'))
		adinfo.whenCreated = lf(d.get('whenCreated'))
		return adinfo


from flask import current_app
from jackdaw.dbmodel.netshare import NetShare

def get_machineid(machineid):
	db = current_app.db
	shares = []
	for share in db.session.query(NetShare).filter_by(machine_id = machineid).all():
		print(share)
		shares.append(share.to_dict())
	
	return shares





from urllib.parse import urlparse

from aiosmb.commons.connection.url import SMBConnectionURL
from msldap.commons.url import MSLDAPURLDecoder

def argchecker(args, param_name, module_name):
	if not hasattr(args, param_name):
		raise Exception('"%s" parameter is mandatory for "%s" operation' % (param_name, module_name))
	if getattr(args, param_name) is None:
		raise Exception('"%s" parameter is mandatory for "%s" operation' % (param_name, module_name))

def construct_smbdef(args):
	return SMBConnectionURL(args.smb_url) #SMBConnectionManager(args.smb_credential_string, proxy_connection_string = args.sproxy)

def construct_ldapdef(args):
	ldap_url = args.ldap_url
	if ldap_url[-1] == '/':
		ldap_url = args.ldap_url[:-1]
	if hasattr(args, 'same_query') and args.same_query is True and args.smb_url is not None:
		ldap_url = '%s/?%s' % (ldap_url, urlparse(args.smb_url).query)
	return MSLDAPURLDecoder(ldap_url)



import os
import connexion
#ORDER IS IMPORTANT!!
from flask_sqlalchemy import SQLAlchemy
#from flask_marshmallow import Marshmallow

from jackdaw.nest.utils.encoder import UniversalFlaskEncoder

basedir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'api'))

# Create the connexion application instance
connex_app = connexion.App(__name__, specification_dir=basedir)
# Read the swagger.yml file to configure the endpoints
connex_app.add_api("swagger.yaml")

# Get the underlying Flask app instance
app = connex_app.app
#set custom JSON encoder
app.json_encoder = UniversalFlaskEncoder

app.config['SQLALCHEMY_ECHO'] = True
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:////home/devel/Desktop/test_hist.db'
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
#
## Create the SQLAlchemy db instance
db = SQLAlchemy(app)

with connex_app.app.app_context():
    connex_app.app.db = db

#
## Initialize Marshmallow
#ma = Marshmallow(app)



from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class NetworkScanJobError(Basemodel):
	__tablename__ = 'networkscanjoberror'
	
	id = Column(Integer, primary_key=True)
	job_id = Column(Integer, ForeignKey('networkscanjob.id'))
	created_at = Column(DateTime, default=datetime.datetime.utcnow)
	error = Column(String, index=True)


	



from flask_sqlalchemy import SQLAlchemy
from jackdaw.dbmodel.adinfo import JackDawADInfo
from flask import current_app

def list_domains():
    db = current_app.db
    domains = []
    for did, distinguishedName, creation in db.session.query(JackDawADInfo).with_entities(JackDawADInfo.id, JackDawADInfo.distinguishedName, JackDawADInfo.fetched_at).all():
        name = distinguishedName.replace('DC=','')
        name = name.replace(',','.')
        domains.append((did, name, creation))
    return domains

def get(domainid):
    db = current_app.db
    adinfo = db.session.query(JackDawADInfo).get(domainid)
    return adinfo.to_dict()



from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class JackDawCustomRelations(Basemodel):
	"""
	This table filled manually by the user(s), and is used to add additional relations between arbitrary nodes (be that user, group, machine etc), ones that could not be determined programmatically
	Data here will be used in node2node path calculations
	"""
	__tablename__ = 'customrelations'
	
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="customrelations", lazy = True)
	created_at = Column(DateTime, default=datetime.datetime.utcnow)
	sid = Column(String, index=True)
	target_sid = Column(String, index=True)


import datetime
import logging

from sqlalchemy.engine import Engine
from sqlalchemy import event
from sqlalchemy import create_engine, Column, Integer, String, DateTime, Boolean, ForeignKey, Table, Enum
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy import create_engine, func
from sqlalchemy.orm import sessionmaker
import enum

from msldap.wintypes.security_descriptor import *

Basemodel = declarative_base()



@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
	cursor = dbapi_connection.cursor()
	cursor.execute("PRAGMA journal_mode = MEMORY")
	cursor.execute("PRAGMA synchronous = OFF")
	cursor.execute("PRAGMA temp_store = MEMORY")
	cursor.execute("PRAGMA cache_size = 500000")
	cursor.close()

def lf(x, sep = ','):
	"""
	flattens objects
	"""
	if isinstance(x, list):
		return sep.join(x)
	elif isinstance(x, (datetime.datetime, int, enum.IntFlag)):
		return x
	return str(x)
	
def dt(x):
	"""
	datetime corrections
	"""
	if x == '':
		return None
	if isinstance(x, str):
		return datetime.datetime.fromisoformat(x)
	if not isinstance(x,datetime.datetime):
		print(x)
	return x
	
def bc(x):
	"""
	boolean corrections
	"""
	if x is None:
		return None
	if isinstance(x,bool):
		return x
	if isinstance(x, str):
		if x.upper() == 'TRUE':
			return True
		elif x.upper() == 'FALSE':
			return False
		elif x.upper() == 'NONE':
			return None
	raise Exception('Cant convert this to bool: %s type: %s' % (x, type(x)))

from .addacl import *
from .adgroup import *
from .adcomp import *
from .adinfo import *
from .aduser import *
from .adou import *
from .credential import *
from .hashentry import *
from .netsession import *
from .netshare import *
from .spnservice import *
from .tokengroup import *
from .usergroup import *
from .localgroup import *
from .constrained import *
from .customrelations import *


def create_db(connection, verbosity = 0):
	logging.info('Creating database %s' % connection)
	engine = create_engine(connection, echo=True if verbosity > 1 else False) #'sqlite:///dump.db'	
	Basemodel.metadata.create_all(engine)
	logging.info('Done creating database %s' % connection)

def get_session(connection, verbosity = 0):
	logging.debug('Connecting to DB')
	engine = create_engine(connection, echo=True if verbosity > 1 else False) #'sqlite:///dump.db'	
	logging.debug('Creating session')
	# create a configured "Session" class
	Session = sessionmaker(bind=engine)
	# create a Session
	return Session()
	
	
am_lookup_table = {
			ADS_ACCESS_MASK.CREATE_CHILD : 'ace_mask_create_child',
			ADS_ACCESS_MASK.DELETE_CHILD : 'ace_mask_delete_child',
			ADS_ACCESS_MASK.ACTRL_DS_LIST : 'ace_mask_actrl_ds_list',
			ADS_ACCESS_MASK.SELF : 'ace_mask_self',
			ADS_ACCESS_MASK.READ_PROP : 'ace_mask_read_prop',
			ADS_ACCESS_MASK.WRITE_PROP : 'ace_mask_write_prop',
			ADS_ACCESS_MASK.DELETE_TREE : 'ace_mask_delete_tree',
			ADS_ACCESS_MASK.LIST_OBJECT : 'ace_mask_list_object',
			ADS_ACCESS_MASK.CONTROL_ACCESS : 'ace_mask_control_access',
			ADS_ACCESS_MASK.DELETE : 'ace_mask_delete',
			ADS_ACCESS_MASK.READ_CONTROL : 'ace_mask_read_control',
			ADS_ACCESS_MASK.WRITE_DACL : 'ace_mask_write_dacl',
			ADS_ACCESS_MASK.WRITE_OWNER : 'ace_mask_write_owner',
			ADS_ACCESS_MASK.SYNCHRONIZE : 'ace_mask_synchronize',
			ADS_ACCESS_MASK.ACCESS_SYSTEM_SECURITY : 'ace_mask_access_system_security',
			ADS_ACCESS_MASK.MAXIMUM_ALLOWED : 'ace_mask_maximum_allowed',
			ADS_ACCESS_MASK.GENERIC_ALL : 'ace_mask_generic_all',
			ADS_ACCESS_MASK.GENERIC_EXECUTE : 'ace_mask_generic_execute',
			ADS_ACCESS_MASK.GENERIC_WRITE : 'ace_mask_generic_write',
			ADS_ACCESS_MASK.GENERIC_READ : 'ace_mask_generic_read',
		}




import sys
import logging

from sqlalchemy import exc

from aiosmb import logger as smblogger
from msldap import logger as msldaplogger

from jackdaw.dbmodel import create_db, get_session
from jackdaw.common.apq import AsyncProcessQueue
from jackdaw.common.proxy import ProxyConnection
from jackdaw.gatherer.universal.smb import SMBGathererManager
#from jackdaw.representation.membership_graph import *

from jackdaw import logger as jdlogger
from jackdaw.gatherer.ldap import LDAPEnumerator
from jackdaw.utils.argshelper import *
from jackdaw.credentials.credentials import JackDawCredentials


def run(args):
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
		jdlogger.setLevel(logging.INFO)
		msldaplogger.setLevel(logging.WARNING)
		smblogger.setLevel(logging.CRITICAL)
		
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
		jdlogger.setLevel(logging.DEBUG)
		msldaplogger.setLevel(logging.INFO)
		smblogger.setLevel(logging.INFO)
		
	elif args.verbose > 1:
		logging.basicConfig(level=1)
		msldaplogger.setLevel(logging.DEBUG)
		jdlogger.setLevel(1)
		smblogger.setLevel(1)
	
	if not args.sql:
		print('SQL connection identification is missing! You need to provide the --sql parameter')
		sys.exit()
	
	db_conn = args.sql
	
	if args.command == 'enum':
		smb_mgr = construct_smbdef(args)
		ldap_mgr = construct_ldapdef(args)

		ldap_conn = ldap_mgr.get_connection()
		ldap_conn.connect()
	
		ldapenum = LDAPEnumerator(db_conn, ldap_conn)
		adifo_id = ldapenum.run()
		print('ADInfo entry successfully created with ID %s' % adifo_id)
		
		mgr = SMBGathererManager(smb_mgr)
		mgr.gathering_type = ['all']
		mgr.db_conn = db_conn
		mgr.target_ad = adifo_id
		mgr.run()
	
	elif args.command == 'dbinit':
		create_db(db_conn)
	
	elif args.command == 'adinfo':
		session = get_session(db_conn)
		from jackdaw.dbmodel.adinfo import JackDawADInfo
		from jackdaw.utils.table import print_table
		
		rows = [['Ad ID', 'domain name', 'scantime']]
		for did, distinguishedName, creation in session.query(JackDawADInfo).with_entities(JackDawADInfo.id, JackDawADInfo.distinguishedName, JackDawADInfo.fetched_at).all():
			name = distinguishedName.replace('DC=','')
			name = name.replace(',','.')
			rows.append([str(did), name, creation.isoformat()])
		print_table(rows)
		
	elif args.command == 'ldap':
		ldap_mgr = construct_ldapdef(args)
		ldap_conn = ldap_mgr.get_connection()
		ldap_conn.connect()
	
		ldapenum = LDAPEnumerator(db_conn, ldap_conn)
		ldapenum.run()
		
	elif args.command in ['shares', 'sessions', 'localgroups']:
		smb_mgr = construct_smbdef(args)
		mgr = SMBGathererManager(smb_mgr)
		mgr.gathering_type = [args.command]
		mgr.db_conn = db_conn
		mgr.lookup_ad = args.lookup_ad
		
		if args.ldap_url:
			ldap_mgr = construct_ldapdef(args)
			ldap_conn = ldap_mgr.get_connection()
			ldap_conn.connect()
			mgr.ldap_conn = ldap_conn
		
		if args.ad_id:
			mgr.target_ad = args.ad_id
		
		if args.target_file:
			mgr.targets_file = args.target_file
		
		mgr.run()
		
	elif args.command == 'creds':
		creds = JackDawCredentials(args.db_conn, args.domain_id)
		creds.add_credentials_impacket(args.impacket_file)

		
	elif args.command == 'passwords':
		creds = JackDawCredentials(args.db_conn)
		creds.add_cracked_passwords(args.potfile, args.disable_usercheck, args.disable_passwordcheck)
		
	elif args.command == 'uncracked':
		creds = JackDawCredentials(args.db_conn, args.domain_id)
		creds.get_uncracked_hashes(args.hash_type, args.history)
			
	elif args.command == 'pwreport':
		creds = JackDawCredentials(args.db_conn, args.domain_id)
		creds.generate_report(args.out_folder)
		
	elif args.command == 'cracked':
		creds = JackDawCredentials(args.db_conn, args.domain_id)
		creds.get_cracked_info()

	elif args.command == 'nest':
		from jackdaw.nest.wrapper import NestServer

		debug = bool(args.verbose)

		server = NestServer(args.sql, bind_ip = args.ip, bind_port = args.port, debug = debug)
		server.run()
	
def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Gather gather gather')
	parser.add_argument('-v', '--verbose', action='count', default=0, help='Increase verbosity, can be stacked')
	parser.add_argument('--sql', help='SQL connection string. When using SQLITE it works best with FULL FILE PATH!!!')
	
	subparsers = parser.add_subparsers(help = 'commands')
	subparsers.required = True
	subparsers.dest = 'command'
	
	nest_group = subparsers.add_parser('nest', formatter_class=argparse.RawDescriptionHelpFormatter, help='Start the Nest server')
	nest_group.add_argument('--ip',  default = '127.0.0.1', help='IP address to listen on')
	nest_group.add_argument('--port',  type=int, default = 5000, help='IP address to listen on')

	adinfo_group = subparsers.add_parser('adinfo', help='Get a list of AD info entries')
	dbinit_group = subparsers.add_parser('dbinit', help='Creates database')
	

	ldap_group = subparsers.add_parser('ldap', formatter_class=argparse.RawDescriptionHelpFormatter, help='Enumerate potentially vulnerable users via LDAP', epilog = MSLDAPURLDecoder.help_epilog)
	ldap_group.add_argument('ldap_url',  help='Connection specitication in URL format')
	
	enum_group = subparsers.add_parser('enum', formatter_class=argparse.RawDescriptionHelpFormatter, help='Enumerate all stuffs', epilog = MSLDAPURLDecoder.help_epilog)
	enum_group.add_argument('ldap_url',  help='Connection specitication in URL format')
	enum_group.add_argument('smb_url',  help='Connection specitication in URL format')
	enum_group.add_argument('-q', '--same-query', action='store_true', help='Use the same query for LDAP as for SMB. LDAP url must still be present, but without a query')
	
	share_group = subparsers.add_parser('shares', help='Enumerate shares on target')
	share_group.add_argument('smb_url',  help='Credential specitication in URL format')
	share_group.add_argument('-t', '--target-file', help='taget file with hostnames. One per line.')
	share_group.add_argument('-l', '--ldap-url', help='ldap_connection_string. Use this to get targets from the domain controller')
	share_group.add_argument('-q', '--same-query', action='store_true', help='Use the same query for LDAP as for SMB. LDAP url must still be present, but without a query')
	share_group.add_argument('-d', '--ad-id', help='ID of the domainfo to poll targets rom the DB')
	share_group.add_argument('-i', '--lookup-ad', help='ID of the domainfo to look up comupter names. Advisable to set for LDAP and file pbased targets')
	
	localgroup_group = subparsers.add_parser('localgroups', help='Enumerate local group memberships on target')
	localgroup_group.add_argument('smb_url',  help='Credential specitication in URL format')
	localgroup_group.add_argument('-t', '--target-file', help='taget file with hostnames. One per line.')
	localgroup_group.add_argument('-l', '--ldap-url', help='ldap_connection_string. Use this to get targets from the domain controller')
	localgroup_group.add_argument('-d', '--ad-id', help='ID of the domainfo to poll targets rom the DB')
	localgroup_group.add_argument('-i', '--lookup-ad', help='ID of the domainfo to look up comupter names. Advisable to set for LDAP and file pbased targets')
	
	session_group = subparsers.add_parser('sessions', help='Enumerate connected sessions on target')
	session_group.add_argument('smb_url',  help='Credential specitication in URL format')
	session_group.add_argument('-t', '--target-file', help='taget file with hostnames. One per line.')
	session_group.add_argument('-l', '--ldap-url', help='ldap_connection_string. Use this to get targets from the domain controller')
	session_group.add_argument('-d', '--ad-id', help='ID of the domainfo to poll targets rom the DB')
	session_group.add_argument('-i', '--lookup-ad', help='ID of the domainfo to look up comupter names. Advisable to set for LDAP and file pbased targets')
	
	credential_group = subparsers.add_parser('creds', help='Add credential information from impacket')
	credential_group.add_argument('impacket_file', help='file with LM and NT hashes, generated by impacket secretsdump.py')
	credential_group.add_argument('-d','--domain-id', type=int, default = -1, help='Domain ID to identify the domain')
	
	passwords_group = subparsers.add_parser('passwords', help='Add password information from hashcat potfile')
	passwords_group.add_argument('potfile', help='hashcat potfile with cracked hashes')
	passwords_group.add_argument('--disable-usercheck', action='store_true', help = 'Disables the user pre-check when inserting to DB. All unique passwords will be uploaded.')
	passwords_group.add_argument('--disable-passwordcheck', action='store_true', help = 'Disables the password uniqueness check. WILL FAIL IF PW IS ALREADY IN THE DB.')
	
	uncracked_group = subparsers.add_parser('uncracked', help='Polls the DB for uncracked passwords')
	uncracked_group.add_argument('-t','--hash-type', default='NT', choices= ['NT', 'LM'])
	uncracked_group.add_argument('--history', action='store_true', help = 'Show password history hashes as well')
	uncracked_group.add_argument('-d','--domain-id', type=int, default = -1, help='Domain ID to identify the domain')
	
	cracked_group = subparsers.add_parser('cracked', help='Polls the DB for cracked passwords')
	cracked_group.add_argument('-d','--domain-id', type=int, default = -1, help='Domain ID to identify the domain')
	
	pwreport_group = subparsers.add_parser('pwreport', help='Generates credential statistics')
	pwreport_group.add_argument('-d','--domain-id', type=int, default = -1, help='Domain ID to identify the domain')
	pwreport_group.add_argument('-o','--out-file', help='Base file name to creates report files in')
	
	args = parser.parse_args()

	run(args)

if __name__ == '__main__':
	main()
	



from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class NetworkScanTarget(Basemodel):
	__tablename__ = 'networkscantarget'
	
	id = Column(Integer, primary_key=True)
	scan_id = Column(Integer, ForeignKey('networkscan.id'))
	created_at = Column(DateTime, default=datetime.datetime.utcnow)
	address = Column(String, index=True)
	protocol = Column(String, index=True)
	port = Column(String, index=True)

	scan = relationship("NetworkScan", back_populates="targets", lazy = True)
	jobs = relationship("NetworkScanJob", back_populates="target", lazy = True)





def print_table(lines, separate_head=True):
	"""Prints a formatted table given a 2 dimensional array"""
	#Count the column width
	widths = []
	for line in lines:
			for i,size in enumerate([len(x) for x in line]):
					while i >= len(widths):
							widths.append(0)
					if size > widths[i]:
							widths[i] = size
	   
	#Generate the format string to pad the columns
	print_string = ""
	for i,width in enumerate(widths):
			print_string += "{" + str(i) + ":" + str(width) + "} | "
	if (len(print_string) == 0):
			return
	print_string = print_string[:-3]
	   
	#Print the actual data
	for i,line in enumerate(lines):
			print(print_string.format(*line))
			if (i == 0 and separate_head):
					print("-"*(sum(widths)+3*(len(widths)-1)))


import copy
#from msldap.core import *
from msldap.core.common import MSLDAPTargetProxy, MSLDAPCredential, MSLDAPTarget, LDAPProxyType, MSLDAPURLDecoder
from msldap.core.connection import MSLDAPConnection
from jackdaw.common.proxy import ProxyConnection, ProxyType
from jackdaw import logger


class LDAPConnectionManager:
	def __init__(self, ldap_connection_string):
		self.ldap_connection_string = ldap_connection_string
		
		self.auth = None
		self.target = None
		self.proxy = None

		self.connection = None
		self.ldapurl = MSLDAPURLDecoder(self.ldap_connection_string)
	
	def get_auth(self):
		if self.auth is None:
			self.auth = self.ldapurl.get_credential() #MSLDAPCredential.from_connection_string(self.ldap_connection_string)
		return copy.deepcopy(self.auth)
		
	def get_target(self):
		if self.target is None:
			self.target = self.ldapurl.get_target()			
		return copy.deepcopy(self.target)

	def get_connection(self):
		if self.connection is None:
			self.get_auth()
			self.get_target()
			logger.debug(self.get_target())
			self.connection = MSLDAPConnection(self.auth, self.target)
			
		return copy.deepcopy(self.connection)



import asyncio
import logging
import json
import traceback
import ipaddress
import multiprocessing
import threading

from tqdm import tqdm
from dns import resolver, reversename

import aiosmb
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.commons.utils.extb import format_exc

from jackdaw.common.apq import AsyncProcessQueue
from jackdaw.dbmodel.netshare import NetShare
from jackdaw.dbmodel.netsession import NetSession
from jackdaw.dbmodel.localgroup import LocalGroup
from jackdaw import logger
from jackdaw.dbmodel import get_session

from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.adcomp import JackDawADMachine

class SMBGathererManager:
	def __init__(self, smb_mgr):
		self.in_q = AsyncProcessQueue()
		self.out_q = AsyncProcessQueue()
		self.smb_mgr = smb_mgr
		self.gathering_type = ['all']
		self.localgroups = ['Administrators', 'Distributed COM Users','Remote Desktop Users']
		self.concurrent_connections = 10
		self.domain = None
		self.dc_ip = None
		self.timeout = 3
		self.db_conn = None

		self.total_targets = 0
		self.targets = []
		self.targets_file = None
		self.ldap_conn = None
		self.target_ad = None
		self.lookup_ad = None #if specified, it will look up the targets in the DB.
		self.out_file = None

		self.gatherer = None
		
		self.use_progress_bar = True
		self.prg_hosts = None
		self.prg_shares = None
		self.prg_sessions = None
		self.prg_groups = None
		self.prg_errors = None

		self.results_thread = None

	def __target_generator(self):
		if self.db_conn is not None:
			session = get_session(self.db_conn)
		
		for target in self.targets:
			tid = -1
			yield (tid, target)

		if self.targets_file is not None:
			tid = -1
			with open(self.targets_file, 'r') as f:
				for line in f:
					line = line.strip()
					yield (tid, line)

		if self.ldap_conn is not None:
			ldap_filter = r'(&(sAMAccountType=805306369))'
			attributes = ['sAMAccountName']
			for entry in self.ldap_conn.pagedsearch(ldap_filter, attributes):
				tid = -1
				if self.lookup_ad is not None:
					res = session.query(JackDawADMachine)\
							.filter_by(ad_id = self.lookup_ad)\
							.with_entities(JackDawADMachine.id)\
							.filter(JackDawADMachine.sAMAccountName == entry['attributes']['sAMAccountName'])\
							.first()
					if res is not None:
						tid = res[0]
				
				yield (tid, entry['attributes']['sAMAccountName'][:-1])

		if self.target_ad is not None:
			for target_id, target_name in session.query(JackDawADMachine).filter_by(ad_id = self.target_ad).with_entities(JackDawADMachine.id, JackDawADMachine.sAMAccountName):
				yield (target_id, target_name[:-1])

		if self.db_conn is not None:
			session.close()

	def get_results(self):
		session = None
		if self.db_conn is not None:
			session = get_session(self.db_conn)
		
		while True:
			x = self.out_q.get()
			if x is None:
				break

			tid, target, result, error = x
			if result is None and error is not None:
				#something went error
				logger.debug('[AIOSMBScanner][TargetError][%s] %s' % (target.get_ip(), error))
				if self.use_progress_bar is True:
					self.prg_errors.update()

			if result is not None:
				if self.use_progress_bar is True:
					if isinstance(result, NetSession):
						self.prg_sessions.update()
					elif isinstance(result, NetShare):
						self.prg_shares.update()
					elif isinstance(result, LocalGroup):
						self.prg_groups.update()

				if session is None:
					logger.debug(target, str(result), error)
				else:
					session.add(result)
					session.commit()

			if result is None and error is None:
				logger.debug('Finished: %s' % target.ip)
				if self.use_progress_bar is True:
					self.prg_hosts.update()
	
	def run(self):
		self.in_q = AsyncProcessQueue()
		self.out_q = AsyncProcessQueue()
		if self.use_progress_bar is True:
			self.prg_hosts = tqdm(desc='HOSTS', ascii = True)
			self.prg_shares = tqdm(desc='Shares', ascii = True)
			self.prg_sessions = tqdm(desc='Sessions', ascii = True)
			self.prg_groups = tqdm(desc='LocalGroup', ascii = True)
			self.prg_errors = tqdm(desc='Errors', ascii = True)

		self.results_thread = threading.Thread(target = self.get_results)
		self.results_thread.daemon = True
		self.results_thread.start()

		self.gatherer = AIOSMBGatherer(self.in_q, self.out_q, self.smb_mgr, gather = self.gathering_type, localgroups = self.localgroups, concurrent_connections = self.concurrent_connections)
		self.gatherer.start()
		
		for target in self.__target_generator():
			self.total_targets += 1
			if self.use_progress_bar is True:
				self.prg_hosts.total = self.total_targets
			self.in_q.put(target)
		
		self.in_q.put(None)
		#if self.use_progress_bar is True:
		#	self.prg_hosts.total = self.total_targets

		self.results_thread.join()


class AIOSMBGatherer(multiprocessing.Process):
	def __init__(self, in_q, out_q, smb_mgr, gather = ['all'], localgroups = [], concurrent_connections = 10):
		multiprocessing.Process.__init__(self)
		self.in_q = in_q
		self.out_q = out_q
		self.smb_mgr = smb_mgr
		self.gather = gather
		self.localgroups = localgroups
		self.concurrent_connections = concurrent_connections

		self.targets = []
		self.worker_q = None

	def setup(self):
		pass

	async def scan_host(self, atarget):
		try:
			tid, target = atarget
			#spneg = AuthenticatorBuilder.to_spnego_cred(self.credential, target)
			connection = self.smb_mgr.create_connection_newtarget(target)
			async with connection:
				await connection.login()
				machine = SMBMachine(connection)

				if 'all' in self.gather or 'shares' in self.gather:
					async for smbshare, err in machine.list_shares():
						if err is not None:
							await self.out_q.coro_put((tid, connection.target, None, 'Failed to list shares. Reason: %s' % format_exc(err)))
							continue
						share = NetShare()
						share.machine_id = tid
						share.ip = connection.target.get_ip()
						share.netname = smbshare.name
						share.type = smbshare.type
						share.remark = smbshare.remark

						await self.out_q.coro_put((tid, connection.target, share, None))
					
				
				if 'all' in self.gather or 'sessions' in self.gather:
					async for session, err in machine.list_sessions():
						if err is not None:
							await self.out_q.coro_put((tid, connection.target, None, 'Failed to get sessions. Reason: %s' % format_exc(err)))
							continue

						sess = NetSession()
						sess.machine_id = tid
						sess.source = connection.target.get_ip()
						sess.ip = session.ip_addr.replace('\\','').strip()
						sess.username = session.username

						await self.out_q.coro_put((tid, connection.target, sess, None))

				if 'all' in self.gather or 'localgroups' in self.gather:
					for group_name in self.localgroups:
						async for domain_name, user_name, sid, err in machine.list_group_members('Builtin', group_name):
							if err is not None:
								await self.out_q.coro_put((tid, connection.target, None, 'Failed to connect to poll group memeberships. Reason: %s' % format_exc(err)))
								continue

							lg = LocalGroup()
							lg.machine_id = tid
							lg.ip = connection.target.get_ip()
							lg.hostname = connection.target.get_hostname()
							lg.sid = sid
							lg.groupname = group_name
							lg.domain = domain_name
							lg.username = user_name
							await self.out_q.coro_put((tid, connection.target, lg, None))
		
		except Exception as e:
			await self.out_q.coro_put((tid, connection.target, None, 'Failed to connect to host. Reason: %s' % format_exc(e)))
			return

		finally:
			await self.out_q.coro_put((tid, connection.target, None, None)) #target finished

	async def worker(self):
		while True:
			try:
				target = await self.worker_q.get()
				if target is None:
					return
				try:
					await self.scan_host(target)
				except:
					#exception should be handled in scan_host
					continue
			except Exception as e:
				logger.exception('WORKER ERROR')
				raise

	async def scan_queue(self):
		"""
		Reads targets from queue and scans them
		"""
		self.worker_q = asyncio.Queue()
		tasks = []
		for _ in range(self.concurrent_connections):
			tasks.append(asyncio.create_task(self.worker()))

		while True:
			target = await self.in_q.coro_get()
			if target is None:
				for _ in range(self.concurrent_connections):
					await self.worker_q.put(None)
				break
			else:
				await self.worker_q.put(target)

		results = await asyncio.gather(*tasks, return_exceptions = True)
		for res in results:
			if isinstance(res, Exception):
				logger.error('Error! %s' % res)
		await self.out_q.coro_put(None)
		

	
	def run(self):
		self.setup()
		try:
			loop = asyncio.get_event_loop()
		except:
			loop = asyncio.new_event_loop()
		#loop.set_debug(True)  # Enable debug
		loop.run_until_complete(self.scan_queue())





import os
import connexion
#ORDER IS IMPORTANT!!
from flask_sqlalchemy import SQLAlchemy
#from flask_marshmallow import Marshmallow

from jackdaw.nest.utils.encoder import UniversalFlaskEncoder

import connexion
import flask
from connexion.apps.flask_app import FlaskJSONEncoder

class ConfigurableFlaskApp(connexion.FlaskApp):
	def __init__(self, import_name, **kwargs):
		self.flask_args = _get_flask_args(kwargs)
		connexion_args = _get_connexion_args(kwargs)
		super().__init__(import_name, **connexion_args)

	def create_app(self):
		app = flask.Flask(self.import_name, **self.flask_args)
		app.json_encoder = FlaskJSONEncoder
		return app

def _get_flask_args(kwargs):
	return {k.replace('flask_', ''): v for k, v in kwargs.items() if k.startswith('flask_')}

def _get_connexion_args(kwargs):
	return {k: v for k, v in kwargs.items() if not k.startswith('flask_')}

class NestServer:
	def __init__(self, db_conn, bind_ip = '127.0.0.1', bind_port = 5000, debug = True, basedir = None, swagger_config = "swagger.yaml"):
		self.basedir = basedir
		self.db_conn_string = db_conn #connection string
		self.swagger_config = swagger_config
		self.bind_ip = bind_ip
		self.bind_port = bind_port
		self.debug = debug

		self.connex_app = None

	def setup(self):
		if self.basedir is None:
			self.basedir = os.path.abspath(os.path.dirname(__file__))

		#print(self.basedir)
		#print(self.db_conn_string)

		# Create the connexion application instance
		self.connex_app = ConfigurableFlaskApp(__name__, flask_instance_path =self.basedir, specification_dir='api')
		# Read the swagger.yml file to configure the endpoints
		self.connex_app.add_api(self.swagger_config)

		# Get the underlying Flask app instance
		app = self.connex_app.app
		#set custom JSON encoder
		app.json_encoder = UniversalFlaskEncoder

		app.config['SQLALCHEMY_ECHO'] = True
		app.config['SQLALCHEMY_DATABASE_URI'] = self.db_conn_string
		app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
		#
		## Create the SQLAlchemy db instance
		db = SQLAlchemy(app)

		with self.connex_app.app.app_context():
			self.connex_app.app.db = db
		#
		## Initialize Marshmallow
		#ma = Marshmallow(app)

	def run(self):
		self.setup()
		self.connex_app.run(host=self.bind_ip, port=self.bind_port, debug=self.debug)






from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.tokengroup import JackDawTokenGroup
from jackdaw.dbmodel import *
from jackdaw.wintypes.well_known_sids import get_name_or_sid, get_sid_for_name
from jackdaw.wintypes.lookup_tables import *
from sqlalchemy import not_, and_, or_, case
from sqlalchemy.orm import load_only
#from pyvis.network import Network
#from pyvis.options import Layout
import networkx as nx
import math
from jackdaw import logger

class GraphNode:
	def __init__(self, gid, name, gtype = None, properties = {}):
		self.name = name
		self.id = gid
		self.type = gtype
		self.properties = properties
		self.mindistance = math.inf

	def set_distance(self, d):
		self.mindistance = min(self.mindistance, d)

	def serialize_mindistance(self):
		if self.mindistance == math.inf:
			return 999999

	def to_dict(self, format = None):
		if format is None:
			return {
				'id' : self.id,
				'name' : self.name,
				'properties' : self.properties,
				'md' : self.serialize_mindistance(),
			}

		elif format == 'd3':
			return {
				'id' : self.id,
				'name' : self.name,
				'type' : self.type,
				'md' : self.serialize_mindistance(),
			}
		
		elif format == 'vis':
			return {
				'id' : self.id,
				'label' : self.name,
				'type' : self.type,
				'md' : self.serialize_mindistance(),
			}

class GraphEdge:
	def __init__(self, src, dst, label = '', weight = 1, properties = {}):
		self.src = src
		self.dst = dst
		self.label = label
		self.weight = weight
		self.properties = properties

	def to_dict(self, format = None):
		if format is None:
			return {
				'src' : self.src,
				'dst' : self.dst,
				'label' : self.label,
				'weight' : self.weight,
				'properties' : self.properties
			}
		elif format == 'd3':
			return {
				'source' : self.src,
				'target' : self.dst,
				'label'  : self.label,
				'weight' : self.weight,
			}
		elif format == 'vis':
			return {
				'from' : self.src,
				'to' : self.dst,
				'label'  : self.label,
				'weight' : self.weight,
			}


class GraphData:
	def __init__(self):
		self.nodes = {}
		self.edges = []

	def add_node(self, gid, name, node_type, properties = {}):
		self.nodes[gid] = GraphNode(gid, name, node_type, properties)
	
	def add_edge(self, src, dst, label = '', weight = 1, properties = {}):
		if src not in self.nodes:
			raise Exception('Node with id %s is not present' % src)
		if dst not in self.nodes:
			raise Exception('Node with id %s is not present' % dst)

		self.edges.append(GraphEdge(src, dst, label, weight, properties))

	def __add__(self, o):
		if not isinstance(o, GraphData):
			raise Exception('Cannot add GraphData and %s' % type(o))
		
		self.nodes.update(o.nodes)
		self.edges += o.edges
		return self

	def to_dict(self, format = None):
		if format is None:
			return {
				'nodes' : [self.nodes[x].to_dict() for x in self.nodes],
				'edges' : [x.to_dict() for x in self.edges]
			}
		elif format == 'd3':
			return {
				'nodes' : [self.nodes[x].to_dict(format = format) for x in self.nodes],
				'links' : [x.to_dict(format = format) for x in self.edges]
			}
		elif format == 'vis':
			return {
				'nodes' : [self.nodes[x].to_dict(format = format) for x in self.nodes],
				'edges' : [x.to_dict(format = format) for x in self.edges]
			}

def ace_applies(ace_guid, object_class):
	'''
	Checks if an ACE applies to this object (based on object classes).
	Note that this function assumes you already verified that InheritedObjectType is set (via the flag).
	If this is not set, the ACE applies to all object types.
	'''
	try:
		our_ace_guid = OBJECTTYPE_GUID_MAP[object_class]
	except KeyError:
		return False
	if ace_guid == our_ace_guid:
		return True
	# If none of these match, the ACE does not apply to this object
	return False


class DomainGraph:
	def __init__(self, db_conn = None, dbsession = None):
		self.db_conn = db_conn
		self.dbsession = dbsession
		self.ad_id = None
		self.graph = nx.DiGraph()
		#self.network_visual = Network("3000px", "3000px")
		self.show_group_memberships = True
		self.show_user_memberships = True
		self.show_machine_memberships = True
		self.show_session_memberships = True
		self.show_localgroup_memberships = False
		self.show_constrained_delegations = True
		self.show_unconstrained_delegations = True
		self.show_custom_relations = True
		self.show_acl = True
		self.unknown_node_color = "#ffffff"
		self.domain_sid = None
		
		self.blacklist_sids = {'S-1-5-32-545': ''}
		self.ignoresids = {"S-1-3-0": '', "S-1-5-18": ''}

	def get_session(self):
		if self.db_conn is not None:
			return get_session(self.db_conn)
		elif self.dbsession is not None:
			return self.dbsession
		else:
			raise Exception('Either db_conn or dbsession MUST be supplied!')
		
	def is_blacklisted_sid(self, sid):
		if sid in self.blacklist_sids:
			return True
		if sid[:len('S-1-5-21')] == 'S-1-5-21':
			if sid[-3:] == '513':
				return True
			
		return False
	
	def add_sid_to_node(self, node, node_type, name = None):
		if self.is_blacklisted_sid(node):
			return
		if not name:
			name = str(get_name_or_sid(str(node)))
		
		#this presence-filter is important, as we will encounter nodes that are known and present in the graph already
		#ald later will be added via tokengroups as unknown
		if node not in self.graph.nodes:
			self.graph.add_node(str(node), name=name, node_type=node_type)
		
			
	def add_edge(self, sid_src, sid_dst, label = None, weight = 1):
		if not sid_src or not sid_dst:
			return
		if self.is_blacklisted_sid(sid_src) or self.is_blacklisted_sid(sid_dst):
			return
			
		self.add_sid_to_node(sid_src, 'unknown')
		self.add_sid_to_node(sid_dst, 'unknown')
			
		self.graph.add_edge(sid_src, sid_dst, label = label, weight = weight)
		
		#self.network_visual.add_edge(sid_src, sid_dst, label = label, weight = weight)
			
	def sid2cn(self, sid, throw = False):
		session = self.get_session()
		tsid = session.query(JackDawTokenGroup.cn).filter(JackDawTokenGroup.sid == sid).first()
		#print('sid2cn: %s' % tsid)
		if not tsid:
			t = str(get_name_or_sid(str(sid)))
			if t == sid and throw == True:
				raise Exception('No CN found for SID = %s' % repr(sid))
			return t
		return tsid[0]
	
	def cn2sid(self, cn, throw = False, domain_sid = None):
		sid = get_sid_for_name(cn, domain_sid)
		
		session = self.get_session()
		tsid = session.query(JackDawTokenGroup.sid).filter(JackDawTokenGroup.cn == cn).first()
		#print(tsid)
		if not tsid:
			if throw == True:
				raise Exception('No SID found for CN = %s' % repr(cn))
			return cn
		return tsid[0]
			
	def show_all(self):
		"""
		Returns all nodes and edges from the graph
		You really don't want this on a larger graph, mostly used for testing purposes
		"""
		data = GraphData()
		for sid in self.graph.nodes:
			data.add_node(sid, self.graph.nodes[sid].get('name', self.sid2cn(sid)), self.graph.nodes[sid].get('node_type'))

		for edge in [e for e in self.graph.edges]:
			data.add_edge(edge[0], edge[1])

		return data

	def show_domain_admins(self):
		dst_sid = self.cn2sid('Domain Admins', True, self.domain_sid)
		return self.all_shortest_paths( dst_sid = dst_sid)
			
	def show_all_sources(self, src):
		src_sid = self.cn2sid(src, True, self.domain_sid)
		
		return self.all_shortest_paths(src_sid = src_sid)
	
	def get_node(self, nodeid = None):
		
		if nodeid is None:
			nodes = []
			for gid, props in list(self.graph.nodes(data=True)):
				nodes.append(GraphNode(gid, gid, properties = {}))
			return nodes

		if nodeid not in self.graph.nodes:
			return None
		gid, *t = self.graph.nodes[nodeid]
		return GraphNode(gid, gid, properties = {})


	def show_all_destinations(self, dst):
		dst_sid = self.cn2sid(dst, True, self.domain_sid)	
		return self.all_shortest_paths( dst_sid = dst_sid)
	
	def show_path(self, src, dst):
		src_sid = self.cn2sid(src, True, self.domain_sid)	
		dst_sid = self.cn2sid(dst, True, self.domain_sid)	

		return self.all_shortest_paths(src_sid = src_sid, dst_sid = dst_sid)
			
	def __add_path(self, network, path):
		"""
		Adds the path to the representational network
		Path is a list of sids (nodes), so we need to find the edges matching
		"""
		#print('PATH: %s' % repr(path))
		for d, sid in enumerate(path):
			#print(sid)
			network.add_node(sid, name = self.graph.nodes[sid].get('name', self.sid2cn(sid)), node_type = self.graph.nodes[sid].get('node_type'))
			network.nodes[sid].set_distance(d)
		
		for i in range(len(path) - 1):
			for edge in self.graph.edges([path[i], ], data=True):
				if edge[1] == path[i + 1]:
					name = edge[2].get('label', None)
					network.add_edge(path[i], path[i + 1], label=name)

	def distances_from_node(self, dst_sid):
		
		distances = {} #distance -> occurrence of distance
		for node in self.graph.nodes:
			try:
				for path in nx.all_shortest_paths(self.graph, source = node, target = dst_sid):
					distance = len(path)
					if distance not in distances:
						distances[distance] = 0
					distances[distance] += 1
			except nx.exception.NetworkXNoPath:
				continue

		return distances

	def all_shortest_paths(self, src_sid = None, dst_sid = None):
		nv = GraphData()
		
		if not src_sid and not dst_sid:
			raise Exception('Either source or destination MUST be specified')
		
		elif not src_sid and dst_sid:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			for node in self.graph.nodes:
				if node == dst_sid:
					continue
				try:
					for path in nx.all_shortest_paths(self.graph, source = node, target = dst_sid):
						self.__add_path(nv, path)
									
				except nx.exception.NetworkXNoPath:
					continue
			
		elif src_sid and not dst_sid:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			
			for node in self.graph.nodes:
				if node == src_sid:
					continue
				try:
					for path in nx.all_shortest_paths(self.graph, source = src_sid, target = node):
						self.__add_path(nv, path)
									
				except nx.exception.NetworkXNoPath:
					continue
					
		else:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			for path in nx.all_shortest_paths(self.graph, source = src_sid, target = dst_sid):
				self.__add_path(nv, path)
		
		return nv
		
	#def plot(self, network):
	#	"""
	#	Creates and opens an HTML file representing the network
	#	
	#	network: pyvis network
	#	"""
	#	layout = Layout()
	#	layout.hierarchical.sortMethod = 'directed'
	#	layout.hierarchical.edgeMinimization = False
	#	layout.hierarchical.blockShifting = False
	#	layout.hierarchical.levelSeparation = 325
	#	layout.hierarchical.enabled = True
	#	layout.hierarchical.nodeSpacing = 325
	#	layout.hierarchical.treeSpacing = 250
	#	layout.hierarchical.direction = 'LR'
	#	network.options.layout = layout
	#	network.show_buttons(filter_=['layout'])
	#	network.show("test.html")
		
	def calc_acl_edges_sql(self, session, ad_id):
		#enumerating owners
		query = session.query(JackDawADDACL.owner_sid, JackDawADDACL.sid).filter(~JackDawADDACL.owner_sid.in_(["S-1-3-0", "S-1-5-18"])).filter(JackDawADDACL.ad_id == ad_id)
		for owner_sid, sid in query.all():
			self.add_edge(owner_sid, sid, label='Owner')
		
		#queriing generic access
		query = session.query(JackDawADDACL)\
						.filter(JackDawADDACL.ace_type == 'ACCESS_ALLOWED_ACE_TYPE')\
						.filter(~JackDawADDACL.ace_sid.in_(["S-1-3-0", "S-1-5-18"]))\
						.filter(JackDawADDACL.ad_id == ad_id)
		#print('ACCESS_ALLOWED_ACE_TYPE')
		for acl in query.all():
			if acl.ace_mask_generic_all == True:
				self.add_edge(acl.ace_sid, acl.sid, label='GenericALL')
			
			if acl.ace_mask_generic_write == True:
				self.add_edge(acl.ace_sid, acl.sid, label='GenericWrite')
				
			if acl.ace_mask_write_owner == True:
				self.add_edge(acl.ace_sid, acl.sid, label='WriteOwner')
				
			if acl.ace_mask_write_dacl == True:
				self.add_edge(acl.ace_sid, acl.sid, label='WriteDacl')
				
			if acl.object_type in ['user', 'domain'] and acl.ace_mask_control_access == True:
				self.add_edge(acl.ace_sid, acl.sid, label='ExtendedRightALL')
		
		#queriing only necessary fileds
		fields = ['ace_mask_generic_all', 'ace_mask_write_dacl', 'ace_mask_write_owner', 'ace_mask_generic_write', 'ace_objecttype', 'object_type', 'ace_mask_write_prop', 'ace_mask_control_access']
		#queriing object type access
		query = session.query(JackDawADDACL)\
						.options(load_only(*fields))\
						.filter(JackDawADDACL.ace_type == 'ACCESS_ALLOWED_OBJECT_ACE_TYPE')\
						.filter(JackDawADDACL.ad_id == ad_id)\
						.filter(~JackDawADDACL.ace_sid.in_(["S-1-3-0", "S-1-5-18"]))\
						.filter(~and_(JackDawADDACL.ace_hdr_flag_inherited == True, JackDawADDACL.ace_hdr_flag_inherit_only == True))\
						.filter(or_(JackDawADDACL.ace_hdr_flag_inherited == False,\
									JackDawADDACL.ace_hdr_flag_inherit_only == False,\
									and_(JackDawADDACL.ace_hdr_flag_inherited == True, JackDawADDACL.ace_hdr_flag_inherit_only == True, JackDawADDACL.ace_inheritedobjecttype == JackDawADDACL.object_type_guid)))
								

		#print('ACCESS_ALLOWED_OBJECT_ACE_TYPE')
		for acl in query.all():			
			if any([acl.ace_mask_generic_all, acl.ace_mask_write_dacl, acl.ace_mask_write_owner, acl.ace_mask_generic_write]):
				if acl.ace_objecttype is not None and not ace_applies(acl.ace_objecttype, acl.object_type):
					continue
				
				if acl.ace_mask_generic_all == True:
					self.add_edge(acl.ace_sid, acl.sid, label='GenericALL')
					continue
			
				if acl.ace_mask_generic_write == True:
					self.add_edge(acl.ace_sid, acl.sid, label='GenericWrite')
					
					if acl.object_type != 'domain':
						continue
					
				if acl.ace_mask_write_dacl == True:
					self.add_edge(acl.ace_sid, acl.sid, label='WriteDacl')
					
				if acl.ace_mask_write_owner == True:
					self.add_edge(acl.ace_sid, acl.sid, label='WriteOwner')
					
			if acl.ace_mask_write_prop == True:
				if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
					self.add_edge(acl.ace_sid, acl.sid, label='GenericWrite')
					
				if acl.object_type == 'group' and acl.ace_objecttype == 'bf9679c0-0de6-11d0-a285-00aa003049e2':
					self.add_edge(acl.ace_sid, acl.sid, label='AddMember')
					
				
			
			if acl.ace_mask_control_access == True:
				if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
					self.add_edge(acl.ace_sid, acl.sid, label='ExtendedAll')
				
				if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6ad-9c07-11d1-f79f-00c04fc2dcd2':
					# 'Replicating Directory Changes All'
					self.add_edge(acl.ace_sid, acl.sid, label='GetChangesALL')
						
				if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6aa-9c07-11d1-f79f-00c04fc2dcd2':
						# 'Replicating Directory Changes'
						self.add_edge(acl.ace_sid, acl.sid, label='GetChanges')
						
				if acl.object_type == 'user' and acl.ace_objecttype == '00299570-246d-11d0-a768-00aa006e0529':
						# 'Replicating Directory Changes'
						self.add_edge(acl.ace_sid, acl.sid, label='User-Force-Change-Password')
	
	def calc_acl_edges(self, adinfo):
		for acl in adinfo.objectacls:
			##SUPER-IMPORTANT!!!!
			##THE ACCESS RIGHTS CALCULATIONS ARE FLAWED (JUST LIKE IN BLOODHOUND)
			##CORRECT WAY WOULD BE TO TRAVERSE ALL ACES IN A GIVEN ACL IN THE ORDER OF "ACE-ODER" AND CHECK FOR DENY POLICIES AS WELL!!!
			##TODO: CALCULATE THE EFFECTIVE PERMISSIONS IN A CORRECT MANNER :)
			
			##TODO: The DB is designed that most of these calculations can (and should) be offloaded to the DB itself
			##      Currently it's in python bc I'm laze and Dirkjan already wrote the major part
			##      https://github.com/fox-it/BloodHound.py/blob/0d3897ba4cc00818c0fc3e01fa3a91fc42e799e2/bloodhound/enumeration/acls.py
			
			if acl.owner_sid not in self.ignoresids:
				self.add_edge(acl.owner_sid, acl.sid, label='Owner')
				
			if acl.ace_sid in self.ignoresids:
				continue
			
			if acl.ace_type in ['ACCESS_ALLOWED_ACE_TYPE','ACCESS_ALLOWED_OBJECT_ACE_TYPE']:
				if acl.ace_type == 'ACCESS_ALLOWED_ACE_TYPE':
					if acl.ace_mask_generic_all == True:
						self.add_edge(acl.ace_sid, acl.sid, label='GenericALL')
					
					if acl.ace_mask_generic_write == True:
						self.add_edge(acl.ace_sid, acl.sid, label='GenericWrite')
						
					if acl.ace_mask_write_owner == True:
						self.add_edge(acl.ace_sid, acl.sid, label='WriteOwner')
						
					if acl.ace_mask_write_dacl == True:
						self.add_edge(acl.ace_sid, acl.sid, label='WriteDacl')
						
					if acl.object_type in ['user', 'domain'] and acl.ace_mask_control_access == True:
						self.add_edge(acl.ace_sid, acl.sid, label='ExtendedRightALL')
				
				if acl.ace_type == 'ACCESS_ALLOWED_OBJECT_ACE_TYPE':
					if acl.ace_hdr_flag_inherited == True and acl.ace_hdr_flag_inherit_only == True:
						continue
					
					if acl.ace_hdr_flag_inherited == True and acl.ace_inheritedobjecttype is not None:
						if not ace_applies(acl.ace_inheritedobjecttype, acl.object_type):
							continue
					
					if any([acl.ace_mask_generic_all, acl.ace_mask_write_dacl, acl.ace_mask_write_owner, acl.ace_mask_generic_write]):
						if acl.ace_objecttype is not None and not ace_applies(acl.ace_objecttype, acl.object_type):
							continue
						
						if acl.ace_mask_generic_all == True:
							self.add_edge(acl.ace_sid, acl.sid, label='GenericALL')
							continue
				
						if acl.ace_mask_generic_write == True:
							self.add_edge(acl.ace_sid, acl.sid, label='GenericWrite')
							
							if acl.object_type != 'domain':
								continue
							
						if acl.ace_mask_write_dacl == True:
							self.add_edge(acl.ace_sid, acl.sid, label='WriteDacl')
							
						if acl.ace_mask_write_owner == True:
							self.add_edge(acl.ace_sid, acl.sid, label='WriteOwner')
							
					if acl.ace_mask_write_prop == True:
						if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
							self.add_edge(acl.ace_sid, acl.sid, label='GenericWrite')
							
						if acl.object_type == 'group' and acl.ace_objecttype == 'bf9679c0-0de6-11d0-a285-00aa003049e2':
							self.add_edge(acl.ace_sid, acl.sid, label='AddMember')
							
						
				
					if acl.ace_mask_control_access == True:
						if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
							self.add_edge(acl.ace_sid, acl.sid, label='ExtendedAll')
						
						if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6ad-9c07-11d1-f79f-00c04fc2dcd2':
							# 'Replicating Directory Changes All'
							self.add_edge(acl.ace_sid, acl.sid, label='GetChangesALL')
								
						if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6aa-9c07-11d1-f79f-00c04fc2dcd2':
								# 'Replicating Directory Changes'
								self.add_edge(acl.ace_sid, acl.sid, label='GetChanges')
								
						if acl.object_type == 'user' and acl.ace_objecttype == '00299570-246d-11d0-a768-00aa006e0529':
								# 'Replicating Directory Changes'
								self.add_edge(acl.ace_sid, acl.sid, label='User-Force-Change-Password')
		
	def construct(self, ad_id):
		"""
		Fills the network graph from database to memory
		"""
		self.ad_id = ad_id
		session = self.get_session()
		adinfo = session.query(JackDawADInfo).get(ad_id)
		
		self.domain_sid = str(adinfo.objectSid)
		
		node_lables = {}
		node_color_map = []
		
		distinct_filter = {}
		if self.show_group_memberships == True:
			#adding group nodes
			for group in adinfo.groups:
				#if group.sid in distinct_filter:
				#	continue
				#distinct_filter[group.sid] = 1
				self.add_sid_to_node(group.sid, 'group', name=group.name)
		
		#distinct_filter = {}
		if self.show_user_memberships == True:
			#adding user nodes
			for user in adinfo.users:
				self.add_sid_to_node(user.objectSid, 'user', name=user.sAMAccountName)
				
		distinct_filter = {}
		if self.show_machine_memberships == True:
			#adding user nodes
			for user in adinfo.computers:
				#if user.objectSid in distinct_filter:
				#	continue
				#distinct_filter[user.objectSid] = 1
				self.add_sid_to_node(user.objectSid, 'machine', name=user.sAMAccountName)
		
		
		if self.show_session_memberships == True:
			for res in session.query(JackDawADUser.objectSid, JackDawADMachine.objectSid).filter(NetSession.username == JackDawADUser.sAMAccountName).filter(NetSession.source == JackDawADMachine.sAMAccountName).distinct(NetSession.username):
				self.add_edge(res[0], res[1], label='hasSession')
				self.add_edge(res[1], res[0], label='hasSession')
		
		if self.show_localgroup_memberships == True:
			#TODO: maybe create edges based on local username similarities??
			for res in session.query(JackDawADUser.objectSid, JackDawADMachine.objectSid, LocalGroup.groupname).filter(LocalGroup.username == JackDawADUser.sAMAccountName).filter(LocalGroup.hostname == JackDawADMachine.sAMAccountName):
				label = None
				if res[3] == 'Remote Desktop Users':
					label = 'canRDP'
					weight = 1
					
				elif res[3] == 'Distributed COM Users':
					label = 'executeDCOM'
					weight = 1
					
				elif res[3] == 'Administrators':
					label = 'adminTo'
					weight = 1
					
				self.add_edge(res[0], res[1], label=label, weight = weight)
			
		if self.show_constrained_delegations == True:
			pass
			
			
		if self.show_unconstrained_delegations == True:
			pass
			
		if self.show_custom_relations == True:
			for res in adinfo.customrelations:
				self.add_edge(res.sid, res.target_sid)
			
		#print('adding membership edges')
		#adding membership edges
		for tokengroup in adinfo.group_lookups:
			self.add_sid_to_node(tokengroup.sid, 'unknown')
			self.add_sid_to_node(tokengroup.member_sid, 'unknown')
				
			if tokengroup.is_user == True and self.show_user_memberships == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, label='member')
				except AssertionError as e:
					logger.exception()
			elif tokengroup.is_machine == True and self.show_machine_memberships == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, label='member')
				except AssertionError as e:
					logger.exception()
			elif tokengroup.is_group == True and self.show_group_memberships == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, label='member')
				except AssertionError as e:
					logger.exception()
		
		if self.show_acl == True:
			logger.info('Adding ACL edges')
			#adding ACL edges
			self.calc_acl_edges(adinfo)
			#self.calc_acl_edges_sql(session, ad_id)
		else:
			logger.info('Not adding ACL edges')
		
						
				
					


from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class NetworkScanJob(Basemodel):
	__tablename__ = 'networkscanjob'
	
	id = Column(Integer, primary_key=True)
	created_at = Column(DateTime, default=datetime.datetime.utcnow)
	target_id = Column(Integer, ForeignKey('networkscantarget.id'))
	plugin = Column(String, index=True)
	plugin_data = Column(String, index=True)


	target = relationship("NetworkScanTarget", back_populates="jobs", lazy = True)
	results = relationship("NetworkScanJobResult", back_populates="job", lazy = True)



from flask import send_file, send_from_directory

def serve_file(filename):
    response = send_from_directory('site', filename)
    response.direct_passthrough = False
    return response

def serve_favico():
    response = send_from_directory('site/static/img', 'favicon.ico')
    response.direct_passthrough = False
    return response

def serve_js(filename):
    response = send_from_directory('site/static/js', filename)
    response.direct_passthrough = False
    return response

def serve_css(filename):
    response = send_from_directory('site/static/css', filename)
    response.direct_passthrough = False
    return response

def serve_img(filename):
    response = send_from_directory('site/static/img', filename)
    response.direct_passthrough = False
    return response

def serve_home():
    response = send_from_directory('site/templates', 'index.html')
    response.direct_passthrough = False
    return response


from . import Basemodel, lf, dt, bc
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class JackDawADOU(Basemodel):
	__tablename__ = 'adou'

	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="ou", lazy = True)
	
	description = Column(String, index=True)
	dn = Column(String, index=True)
	gPLink = Column(String, index=True)
	name = Column(String, index=True)
	objectCategory = Column(String, index=True)
	objectClass = Column(String, index=True)
	objectGUID = Column(String, index=True)
	ou = Column(String, index=True)
	systemFlags = Column(Integer, index=True)
	whenChanged = Column(String, index=True)
	whenCreated = Column(String, index=True)
	
	def to_dict(self):
		return {
			'id' : self.id ,
			'ad_id' : self.ad_id ,
			'description' : self.description ,
			'guid' : self.objectGUID,
			'dn' : self.dn ,
			'name' : self.name ,
			'ou' : self.ou ,
			'whenChanged' : self.whenChanged ,
			'whenCreated' : self.whenCreated ,	
		}
	
	@staticmethod
	def from_adou(u):
		adou = JackDawADOU()
		adou.description = lf(getattr(u,'description'))
		adou.dn = lf(getattr(u,'distinguishedName'))
		adou.gPLink = lf(getattr(u,'gPLink'))
		adou.name = lf(getattr(u,'name'))
		adou.objectCategory = lf(getattr(u,'objectCategory'))
		adou.objectClass = lf(getattr(u,'objectClass'))
		adou.objectGUID = lf(getattr(u,'objectGUID'))
		adou.ou = lf(getattr(u,'ou'))
		adou.systemFlags = lf(getattr(u,'systemFlags'))
		adou.whenChanged = dt(lf(getattr(u,'whenChanged')))
		adou.whenCreated = dt(lf(getattr(u,'whenCreated')))
			
		return adou
		


from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey

class JackDawSPNService(Basemodel):
	__tablename__ = 'spnservice'
	
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="spnservices", lazy = True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	computername = Column(String, index=True)
	service = Column(String, index=True)
	port = Column(Integer, index=True)


from . import Basemodel, lf
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey

class JackDawADGroup(Basemodel):
	__tablename__ = 'groups'
	
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="groups", lazy = True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	cn = Column(String, index=True)
	dn = Column(String, index=True)
	
	guid = Column(String, index=True)
	sid = Column(String, index=True)
	description = Column(String, index=True)
	grouptype = Column(String, index=True)
	
	instanceType = Column(String, index=True)	
	name = Column(String, index=True)	
	member = Column(String, index=True)	
	sAMAccountName = Column(String, index=True)	
	sAMAccountType = Column(String, index=True)	
	systemFlags = Column(String, index=True)	
	whenChanged = Column(String, index=True)	
	whenCreated = Column(String, index=True)

	def to_dict(self):
		return {
			'id' : self.id ,
			'ad_id' : self.ad_id ,
			'sid' : self.sid ,
			'description' : self.description ,
			'grouptype' : self.grouptype ,
			'name' : self.name ,
			'member' : self.member ,
			'sAMAccountName' : self.sAMAccountName ,
			'systemFlags' : self.systemFlags ,
			'whenChanged' : self.whenChanged ,
			'whenCreated' : self.whenCreated ,
		}

	@staticmethod
	def from_dict(d):
		group = JackDawADGroup()
		group.cn = lf(d.get('cn'))
		group.dn = lf(d.get('distinguishedName'))
		group.guid = lf(d.get('objectGUID'))
		group.sid = lf(d.get('objectSid'))
		group.description = lf(d.get('description'))
		group.grouptype = lf(d.get('groupType'))
		group.instanceType = lf(d.get('instanceType'))
		group.name = lf(d.get('name'))
		group.member = lf(d.get('member'))
		group.sAMAccountName = lf(d.get('sAMAccountName'))
		group.sAMAccountType = lf(d.get('sAMAccountType'))
		group.systemFlags = lf(d.get('systemFlags'))
		group.whenChanged = lf(d.get('whenChanged'))
		group.whenCreated = lf(d.get('whenCreated'))
		return group


from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

from msldap.wintypes.security_descriptor import *


am_lookup_table = {
			ADS_ACCESS_MASK.CREATE_CHILD : 'ace_mask_create_child',
			ADS_ACCESS_MASK.DELETE_CHILD : 'ace_mask_delete_child',
			ADS_ACCESS_MASK.ACTRL_DS_LIST : 'ace_mask_actrl_ds_list',
			ADS_ACCESS_MASK.SELF : 'ace_mask_self',
			ADS_ACCESS_MASK.READ_PROP : 'ace_mask_read_prop',
			ADS_ACCESS_MASK.WRITE_PROP : 'ace_mask_write_prop',
			ADS_ACCESS_MASK.DELETE_TREE : 'ace_mask_delete_tree',
			ADS_ACCESS_MASK.LIST_OBJECT : 'ace_mask_list_object',
			ADS_ACCESS_MASK.CONTROL_ACCESS : 'ace_mask_control_access',
			ADS_ACCESS_MASK.DELETE : 'ace_mask_delete',
			ADS_ACCESS_MASK.READ_CONTROL : 'ace_mask_read_control',
			ADS_ACCESS_MASK.WRITE_DACL : 'ace_mask_write_dacl',
			ADS_ACCESS_MASK.WRITE_OWNER : 'ace_mask_write_owner',
			ADS_ACCESS_MASK.SYNCHRONIZE : 'ace_mask_synchronize',
			ADS_ACCESS_MASK.ACCESS_SYSTEM_SECURITY : 'ace_mask_access_system_security',
			ADS_ACCESS_MASK.MAXIMUM_ALLOWED : 'ace_mask_maximum_allowed',
			ADS_ACCESS_MASK.GENERIC_ALL : 'ace_mask_generic_all',
			ADS_ACCESS_MASK.GENERIC_EXECUTE : 'ace_mask_generic_execute',
			ADS_ACCESS_MASK.GENERIC_WRITE : 'ace_mask_generic_write',
			ADS_ACCESS_MASK.GENERIC_READ : 'ace_mask_generic_read',
		}
		
hdr_flag_lookup = {
	AceFlags.CONTAINER_INHERIT_ACE : 'ace_hdr_flag_container_inherit',
	AceFlags.FAILED_ACCESS_ACE_FLAG : 'ace_hdr_flag_failed_access',
	AceFlags.INHERIT_ONLY_ACE : 'ace_hdr_flag_inherit_only',
	AceFlags.INHERITED_ACE : 'ace_hdr_flag_inherited',
	AceFlags.NO_PROPAGATE_INHERIT_ACE : 'ace_hdr_flag_no_propagate_inherit',
	AceFlags.OBJECT_INHERIT_ACE : 'ace_hdr_flag_object_inherit',
	AceFlags.SUCCESSFUL_ACCESS_ACE_FLAG : 'ace_hdr_flag_successful_access',
}

class JackDawADDACL(Basemodel):
	__tablename__ = 'addacl'
	
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="objectacls", lazy = True)
	
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	guid = Column(String, index=True)
	sid = Column(String, index=True)
	
	object_type = Column(String, index=True)
	object_type_guid = Column(String, index=True)
	owner_sid = Column(String, index=True)
	group_sid = Column(String, index=True)
	
	cn = Column(String, index=True)
	dn = Column(String, index=True)
	
	sd_control = Column(String, index=True)
	ace_type = Column(String, index=True)
	ace_mask = Column(String, index=True)
	ace_objecttype = Column(String, index=True)
	ace_inheritedobjecttype = Column(String, index=True)
	ace_sid = Column(String, index=True)
	ace_order = Column(Integer, index = True)
	
	#ace header flags
	
	ace_hdr_flag_container_inherit    = Column(Boolean, index = True)
	ace_hdr_flag_failed_access        = Column(Boolean, index = True)
	ace_hdr_flag_inherit_only         = Column(Boolean, index = True)
	ace_hdr_flag_inherited            = Column(Boolean, index = True)
	ace_hdr_flag_no_propagate_inherit = Column(Boolean, index = True)
	ace_hdr_flag_object_inherit       = Column(Boolean, index = True)
	ace_hdr_flag_successful_access    = Column(Boolean, index = True)
	
	#storing the bitfield in separate columns for easy lookup
	ace_mask_create_child   = Column(Boolean, index = True)
	ace_mask_delete_child   = Column(Boolean, index = True)
	ace_mask_actrl_ds_list  = Column(Boolean, index = True)
	ace_mask_self           = Column(Boolean, index = True)
	ace_mask_read_prop      = Column(Boolean, index = True)
	ace_mask_write_prop     = Column(Boolean, index = True)
	ace_mask_delete_tree    = Column(Boolean, index = True)
	ace_mask_list_object    = Column(Boolean, index = True)
	ace_mask_control_access = Column(Boolean, index = True)
	ace_mask_generic_read   = Column(Boolean, index = True)
	ace_mask_generic_write  = Column(Boolean, index = True)
	ace_mask_generic_execute = Column(Boolean, index = True)
	ace_mask_generic_all    = Column(Boolean, index = True)
	ace_mask_maximum_allowed = Column(Boolean, index = True)
	ace_mask_access_system_security = Column(Boolean, index = True)
	ace_mask_synchronize     = Column(Boolean, index = True)
	ace_mask_write_owner     = Column(Boolean, index = True)
	ace_mask_write_dacl      = Column(Boolean, index = True)
	ace_mask_read_control    = Column(Boolean, index = True)
	ace_mask_delete          = Column(Boolean, index = True)
	
	def hdrflag2attr(hdrflag):
		true_attrs = []
		false_attrs = []
		for hdr in hdr_flag_lookup:
			if hdrflag & hdr:
				true_attrs.append(hdr_flag_lookup[hdr])
			else:
				false_attrs.append(hdr_flag_lookup[hdr])

		return true_attrs, false_attrs
		
	
	@staticmethod
	def mask2attr(ace_mask):
		true_attrs = []
		false_attrs = []
		for am in am_lookup_table:
			if ace_mask & am:
				true_attrs.append(am_lookup_table[am])
			else:
				false_attrs.append(am_lookup_table[am])

		return true_attrs, false_attrs



#https://support.microsoft.com/en-us/help/243330/well-known-security-identifiers-in-windows-operating-systems

#all of the domain sids start with 'S-1-5-21domain-'
WELL_KNOWN_DOMAIN_SIDS = {
	'500' : 'Administrator', #A user account for the system administrator. By default, it is the only user account that is given full control over the system.
	'501' : 'Guest', #A user account for people who do not have individual accounts. This user account does not require a password. By default, the Guest account is disabled.
	'502' : 'KRBTGT', #A service account that is used by the Key Distribution Center (KDC) service.
	'512' : 'Domain Admins', #A global group whose members are authorized to administer the domain. 
	'513' : 'Domain Users', #A global group whose members are authorized to administer the domain. 
	'514' : 'Domain Guests',
	'515' : 'Domain Computers',
	'516' : 'Domain Controllers',
	'517' : 'Cert Publishers',
	'518' : 'Schema Admins',
	'519' : 'Enterprise Admins',
	'520' : 'Group Policy Creator Owners',
	'526' : 'Key Admins',
	'527' : 'Enterprise Key Admins',
	'553' : 'RAS and IAS Servers',
	'498' : 'Enterprise Read-only Domain Controllers',
	'521' : 'Read-only Domain Controllers',
	'571' : 'Allowed RODC Password Replication Group',
	'572' : 'Denied RODC Password Replication Group',
	'522' : 'Cloneable Domain Controllers',
}

WELL_KNOWN_DOMAIN_SIDS_INV = {v: k for k, v in WELL_KNOWN_DOMAIN_SIDS.items()}

#one extra: 'S-1-5-5-X-Y' : 'Logon Session', #A logon session. The X and Y values for these SIDs are different for each session.
WELL_KNOWN_SIDS = {
	'S-1-0' : 'Null Authority',
	'S-1-0-0' : 'Nobody',
	'S-1-1' : 'World Authority',
	'S-1-1-0' : 'Everyone',
	'S-1-2' : 'Local Authority',
	'S-1-2-0' : 'Local',
	'S-1-2-1' : 'Console Logon',
	'S-1-3' : 'Creator Authority',
	'S-1-3-0' : 'Creator Owner',
	'S-1-3-1' : 'Creator Group',
	'S-1-3-2' : 'Creator Owner Server',
	'S-1-3-3' : 'Creator Group Server',
	'S-1-3-4' : 'Owner Rights',
	'S-1-5-80-0' : 'All Services',
	'S-1-4' : 'Non-unique Authority',
	'S-1-5' : 'NT Authority',
	'S-1-5-1' : 'Dialup',
	'S-1-5-2' : 'Network',
	'S-1-5-3' : 'Batch',
	'S-1-5-4' : 'Interactive',
	'S-1-5-6' : 'Service',
	'S-1-5-7' : 'Anonymous',
	'S-1-5-8' : 'Proxy',
	'S-1-5-9' : 'Enterprise Domain Controllers',
	'S-1-5-10' : 'Principal Self',
	'S-1-5-11' : 'Authenticated Users',
	'S-1-5-12' : 'Restricted Code',
	'S-1-5-13' : 'Terminal Server Users',
	'S-1-5-14' : 'Remote Interactive Logon',
	'S-1-5-15' : 'This Organization',
	'S-1-5-17' : 'This Organization',
	'S-1-5-18' : 'Local System',
	'S-1-5-19' : 'NT Authority',
	'S-1-5-20' : 'NT Authority',
	'S-1-5-32-544' : 'Administrators',
	'S-1-5-32-545' : 'Users',
	'S-1-5-32-546' : 'Guests',
	'S-1-5-32-547' : 'Power Users',
	'S-1-5-32-548' : 'Account Operators',
	'S-1-5-32-549' : 'Server Operators',
	'S-1-5-32-550' : 'Print Operators',
	'S-1-5-32-551' : 'Backup Operators',
	'S-1-5-32-552' : 'Replicators',
	'S-1-5-64-10' : 'NTLM Authentication',
	'S-1-5-64-14' : 'SChannel Authentication',
	'S-1-5-64-21' : 'Digest Authentication',
	'S-1-5-80' : 'NT Service',
	'S-1-5-80-0' : 'NT SERVICES\\ALL SERVICES',
	'S-1-5-83-0' : 'NT VIRTUAL MACHINE\\Virtual Machines',
	'S-1-16-0' : 'Untrusted Mandatory Level',
	'S-1-16-4096' : 'Low Mandatory Level',
	'S-1-16-8192' : 'Medium Mandatory Level',
	'S-1-16-8448' : 'Medium Plus Mandatory Level',
	'S-1-16-12288' : 'High Mandatory Level',
	'S-1-16-16384' : 'System Mandatory Level',
	'S-1-16-20480' : 'Protected Process Mandatory Level',
	'S-1-16-28672' : 'Secure Process Mandatory Level',
	'S-1-5-32-554' : 'BUILTIN\\Pre-Windows 2000 Compatible Access',
	'S-1-5-32-555' : 'BUILTIN\\Remote Desktop Users',
	'S-1-5-32-556' : 'BUILTIN\\Network Configuration Operators',
	'S-1-5-32-557' : 'BUILTIN\\Incoming Forest Trust Builders',
	'S-1-5-32-558' : 'BUILTIN\\Performance Monitor Users',
	'S-1-5-32-559' : 'BUILTIN\\Performance Log Users',
	'S-1-5-32-560' : 'BUILTIN\\Windows Authorization Access Group',
	'S-1-5-32-561' : 'BUILTIN\\Terminal Server License Servers',
	'S-1-5-32-562' : 'BUILTIN\\Distributed COM Users',
	'S-1-5-32-569' : 'BUILTIN\\Cryptographic Operators',
	'S-1-5-32-573' : 'BUILTIN\\Event Log Readers',
	'S-1-5-32-574' : 'BUILTIN\\Certificate Service DCOM Access',
	'S-1-5-32-575' : 'BUILTIN\\RDS Remote Access Servers',
	'S-1-5-32-576' : 'BUILTIN\\RDS Endpoint Servers',
	'S-1-5-32-577' : 'BUILTIN\\RDS Management Servers',
	'S-1-5-32-578' : 'BUILTIN\\Hyper-V Administrators',
	'S-1-5-32-579' : 'BUILTIN\\Access Control Assistance Operators',
	'S-1-5-32-580' : 'BUILTIN\\Remote Management Users',
}

WELL_KNOWN_SIDS_INV = {v: k for k, v in WELL_KNOWN_DOMAIN_SIDS.items()}

def get_sid_for_name(name, domain_sid = None):
	if domain_sid:
		if name in WELL_KNOWN_DOMAIN_SIDS_INV:
			return domain_sid + '-' + WELL_KNOWN_DOMAIN_SIDS_INV[name]
	else:
		if name in WELL_KNOWN_SIDS_INV:
			return WELL_KNOWN_SIDS_INV[name]
			
	return None
	

def get_name_or_sid(sid_str):
	if sid_str[:len('S-1-5-5-')] == 'S-1-5-5-':
		return 'Logon Session'
	if sid_str in WELL_KNOWN_SIDS:
		return WELL_KNOWN_SIDS[sid_str]
	if sid_str[:len('S-1-5-21')] == 'S-1-5-21':
		if sid_str[-3:] in WELL_KNOWN_DOMAIN_SIDS:
			return WELL_KNOWN_DOMAIN_SIDS[sid_str[-3:]]
	return sid_str
			
		


# System modules
from datetime import datetime
from jackdaw.nest.graph.domain import DomainGraph, GraphData

# 3rd party modules
from flask import make_response, abort, current_app
from flask import current_app


graph_id_ctr = 1
graphs = {}

def create(adid):
    global graph_id_ctr
    db = current_app.db
    dg = DomainGraph(dbsession = db.session)
    dg.construct(adid)
    graphs[graph_id_ctr] = dg
    graph_id_ctr += 1
    return {'graphid' : graph_id_ctr - 1}

def delete(graphid):
    return {}

def get(graphid):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    res = graphs[graphid].all_shortest_paths()
    return res.to_dict()

def query_path(graphid, src = None, dst = None, format = 'd3'):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    if src is None and dst is None:
        return {}
    res =  graphs[graphid].all_shortest_paths(src, dst)
    return res.to_dict(format = format)

def query_path_da(graphid, format = 'vis'):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    
    da_sids = {}
    #searching for domain admin SID
    for node in graphs[graphid].get_node():
        if node.id.endswith('-512'):
            da_sids[node.id] = 1
    
    #print(da_sids)
    if len(da_sids) == 0:
        return 'Graph Not Found', 404
    
    res = GraphData()
    for sid in da_sids:
        res += graphs[graphid].all_shortest_paths(None, sid)


    #print(res)
    return res.to_dict(format = format)


def list_nodes(graphid, with_data = False):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    nodes = []
    for node in graphs[graphid].get_node():
        nodes.append(node.to_dict())
    return nodes

def get_node(graphid, nodeid):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    return graphs[graphid].get_node(nodeid)

def query_path_all(graphid):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    return graphs[graphid].show_all().to_dict(format = 'd3')

def search_sid(graphid, sid):
    return {}

def search_cn(graphid, cn):
    return {}

def stat_distance(graphid, sid):
    if graphid not in graphs:
        return 'Graph Not Found', 404
    distances = graphs[graphid].distances_from_node(sid)
    
    return distances

def diff(graphids):
    print(graphids)
    original = graphids['src']
    new = graphids['dst']
    
    users_added = {}
    users_removed = {}

    machines_added = {}
    machines_removed = {}

    groups_added = {}
    groups_removed = {}

    for sid, attrs in graphs[new].graph.nodes(data=True):
        print(sid)
        if not graphs[original].graph.has_node(sid):
            print(attrs)
            if attrs.get('node_type') == 'user':
                users_added[sid] = 1
            elif attrs.get('node_type') == 'machine':
                machines_added[sid] = 1
            elif attrs.get('node_type') == 'group':
                groups_added[sid] = 1
    
    for sid, attrs in graphs[original].graph.nodes(data=True):
        if not graphs[new].graph.has_node(sid):
            print(attrs)
            if attrs.get('node_type') == 'user':
                users_removed[sid] = 1
            elif attrs.get('node_type') == 'machine':
                machines_removed[sid] = 1
            elif attrs.get('node_type') == 'group':
                groups_removed[sid] = 1

    print(users_added)
    print(users_removed)
    print(machines_added)
    print(machines_removed)
    print(groups_added)
    print(groups_removed)

    return {}



from flask import current_app
from jackdaw.dbmodel.adgroup import JackDawADGroup

def list_groups(domainid):
    db = current_app.db
    domains = []
    for uid, sid, name in db.session\
                            .query(JackDawADGroup)\
                                .filter_by(ad_id = domainid)\
                                .with_entities(JackDawADGroup.id, JackDawADGroup.sid, JackDawADGroup.sAMAccountName)\
                                .all():
        domains.append((uid, sid, name))
    return domains

def get(domainid, groupid):
    db = current_app.db
    user = db.session.query(JackDawADGroup).get(groupid)
    return user.to_dict()

def get_sid(domainid, sid):
    db = current_app.db
    for user in db.session.query(JackDawADGroup).filter_by(sid = sid).filter(JackDawADGroup.ad_id == domainid).all():
        return user.to_dict()


import enum
from urllib.parse import urlparse, parse_qs

class ProxyType(enum.Enum):
	SOCKS5 = 'SOCKS5'
	SOCKS5_SSL = 'SOCKS5_SSL'
	MULTIPLEXOR = 'MULTIPLEXOR'
	MULTIPLEXOR_SSL = 'MULTIPLEXOR_SSL'

class ProxyConnection:
	"""
	socks5://127.0.0.1:5555
	socks5://user:password@127.0.0.1:5555
	socks5-ssl://127.0.0.1:4444
	socks5-ssl://user:password@127.0.0.1:5555
	multiplexor://127.0.0.1:5555/agentid
	multiplexor://user:password@127.0.0.1:5555/agentid
	multiplexor-ssl://127.0.0.1:5555/agentid
	multiplexor-ssl://user:password@127.0.0.1:5555/agentid
	"""
	def __init__(self):
		self.ip = None
		self.port = None
		self.username = None
		self.domain = None
		self.password = None
		self.timeout = None
		self.type = None
		self.dns = None

	@staticmethod
	def from_connection_string(s):
		url_e = urlparse(s)
		pt = ProxyType(url_e.scheme.replace('-', '_').upper())
		if pt in [ProxyType.SOCKS5, ProxyType.SOCKS5_SSL]:
			pc = Socks5ProxyConnection()
		elif pt in [ProxyType.MULTIPLEXOR, ProxyType.MULTIPLEXOR_SSL]:
			pc = MultiplexorProxyConnection()
		
		pc.type = pt
		pc.ip = url_e.hostname
		pc.port = url_e.port

		if url_e.username is not None:
			if url_e.username.find('\\') != -1:
				pc.domain, pc.username = url_e.username.split('\\')
			else:
				pc.username = url_e.username
			pc.password = url_e.password

		params = parse_qs(url_e.params)
		if 'timeout' in params:
			pc.timeout = int(params['timeout'][0])
		if 'dns' in params:
			pc.dns = params['dns']
		
		pc.parse_rest(url_e)
		return pc

	def __str__(self):
		t = '==== ProxyConnection ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t


class Socks5ProxyConnection(ProxyConnection):
	def __init__(self):
		ProxyConnection.__init__(self)

	def parse_rest(self, url_e):
		return

	def get_ldap(self):
		pass

	def get_smb(self):
		pass


class MultiplexorProxyConnection(ProxyConnection):
	def __init__(self):
		ProxyConnection.__init__(self)
		self.agentid = None

	def parse_rest(self, url_e):
		self.agentid = url_e.path.replace('/','')
		if self.agentid is None:
			raise Exception('Multiplexor proxy requires agentid to be set!')
		return

	def get_ldap(self):
		pass

	def get_smb(self):
		pass


if __name__ == '__main__':
	url_tests = [
		'socks5://127.0.0.1:5555',
		'socks5://user:password@127.0.0.1:5555',
		'socks5://aaa\\user:password@127.0.0.1:5555',
		'socks5-ssl://127.0.0.1:4444',
		'socks5-ssl://user:password@127.0.0.1:5555',
		'multiplexor://127.0.0.1:5555/agentid',
		'multiplexor://user:password@127.0.0.1:5555/agentid',
		'multiplexor-ssl://127.0.0.1:5555/agentid',
		'multiplexor-ssl://user:password@127.0.0.1:5555/agentid',
		'multiplexor-ssl://aaa\\user:password@127.0.0.1:5555/agentid',
		'multiplexor-ssl://aaa\\user:password@127.0.0.1:5555/agentid?user=alma&domain=test&password=44444',

	]
	for url in url_tests:
		print('===========================================================================')
		print(url)
		try:
			dec = ProxyConnection.from_connection_string(url)
		except Exception as e:
			import traceback
			traceback.print_exc()
			print('ERROR! Reason: %s' % e)
			input()
		else:
			print(str(dec))
			#print(str(target))
			input()


from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class NetworkScanJobResult(Basemodel):
	__tablename__ = 'networkscanjobresult'
	
	id = Column(Integer, primary_key=True)
	job_id = Column(Integer, ForeignKey('networkscanjob.id'))
	created_at = Column(DateTime, default=datetime.datetime.utcnow)
	is_error = Column(Boolean, index=True)
	result = Column(String, index=True)



from . import Basemodel, lf, dt, bc
from jackdaw.dbmodel.utils import *
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean


class JackDawADUser(Basemodel):
	__tablename__ = 'users'

	# Now for the attributes
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="users", lazy = True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	sn = Column(String)
	cn = Column(String)
	dn = Column(String)
	initials = Column(String)
	givenName = Column(String)
	displayName = Column(String)
	name = Column(String)
	objectCategory = Column(String)
	objectClass = Column(String)
	objectGUID = Column(String)
	objectSid = Column(String, index=True)
	primaryGroupID = Column(String)
	sAMAccountName = Column(String, index=True)
	userPrincipalName = Column(String)
	servicePrincipalName = Column(String)
	## groups
	memberOf = Column(String) #list, should be extra table
	member = Column(String) #list, should be extra table
	## times
	accountExpires = Column(DateTime)
	badPasswordTime = Column(DateTime)
	lastLogoff = Column(DateTime)
	lastLogon = Column(DateTime)
	lastLogonTimestamp = Column(DateTime)
	pwdLastSet = Column(DateTime)
	whenChanged = Column(DateTime)
	whenCreated = Column(DateTime)
	## security
	badPwdCount = Column(Integer)
	logonCount = Column(Integer)
	sAMAccountType = Column(Integer)
	userAccountControl = Column(Integer)
	
	## other
	codePage = Column(Integer)
	countryCode = Column(Integer)
	
	## calculated properties
	when_pw_change = Column(DateTime)
	when_pw_expires = Column(DateTime)
	must_change_pw = Column(DateTime)
	canLogon = Column(Boolean)
	isAdmin = Column(Boolean)
	
	UAC_SCRIPT = Column(Boolean)
	UAC_ACCOUNTDISABLE = Column(Boolean)
	UAC_HOMEDIR_REQUIRED = Column(Boolean)
	UAC_LOCKOUT = Column(Boolean)
	UAC_PASSWD_NOTREQD = Column(Boolean)
	UAC_PASSWD_CANT_CHANGE = Column(Boolean)
	UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED = Column(Boolean)
	UAC_TEMP_DUPLICATE_ACCOUNT = Column(Boolean)
	UAC_NORMAL_ACCOUNT = Column(Boolean)
	UAC_INTERDOMAIN_TRUST_ACCOUNT = Column(Boolean)
	UAC_WORKSTATION_TRUST_ACCOUNT = Column(Boolean)
	UAC_SERVER_TRUST_ACCOUNT = Column(Boolean)
	UAC_NA_1 = Column(Boolean)
	UAC_NA_2 = Column(Boolean)
	UAC_DONT_EXPIRE_PASSWD = Column(Boolean)
	UAC_MNS_LOGON_ACCOUNT = Column(Boolean)
	UAC_SMARTCARD_REQUIRED = Column(Boolean)
	UAC_TRUSTED_FOR_DELEGATION = Column(Boolean)
	UAC_NOT_DELEGATED = Column(Boolean)
	UAC_USE_DES_KEY_ONLY = Column(Boolean)
	UAC_DONT_REQUIRE_PREAUTH = Column(Boolean)
	UAC_PASSWORD_EXPIRED = Column(Boolean)
	UAC_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION = Column(Boolean)

	#credential = relationship("Credential", back_populates="user", lazy = True)
	allowedtodelegateto = relationship("JackDawUserConstrainedDelegation", back_populates="user", lazy = True)
	

	def to_dict(self):
		return {
			'id' : self.id ,
			'ad_id' : self.ad_id ,
			'dn' : self.dn ,
			'givenName' : self.givenName ,
			'displayName' : self.displayName ,
			'name' : self.name ,
			'objectSid' : self.objectSid ,
			'sAMAccountName' : self.sAMAccountName ,
			'userPrincipalName' : self.userPrincipalName ,
			'servicePrincipalName' : self.servicePrincipalName ,
			'accountExpires' : self.accountExpires ,
			'badPasswordTime' : self.badPasswordTime ,
			'lastLogoff' : self.lastLogoff ,
			'lastLogon' : self.lastLogon ,
			'lastLogonTimestamp' : self.lastLogonTimestamp ,
			'pwdLastSet' : self.pwdLastSet ,
			'whenChanged' : self.whenChanged ,
			'whenCreated' : self.whenCreated ,
			'badPwdCount' : self.badPwdCount ,
			'logonCount' : self.logonCount ,
			'sAMAccountType' : self.sAMAccountType ,
			'userAccountControl' : self.userAccountControl ,
			'codePage' : self.codePage ,
			'countryCode' : self.countryCode ,
			'when_pw_change' : self.when_pw_change ,
			'when_pw_expires' : self.when_pw_expires ,
			'must_change_pw' : self.must_change_pw ,
			'canLogon' : self.canLogon ,
			'UAC_SCRIPT' : self.UAC_SCRIPT ,
			'UAC_ACCOUNTDISABLE' : self.UAC_ACCOUNTDISABLE ,
			'UAC_HOMEDIR_REQUIRED' : self.UAC_HOMEDIR_REQUIRED ,
			'UAC_LOCKOUT' : self.UAC_LOCKOUT ,
			'UAC_PASSWD_NOTREQD' : self.UAC_PASSWD_NOTREQD ,
			'UAC_PASSWD_CANT_CHANGE' : self.UAC_PASSWD_CANT_CHANGE ,
			'UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED' : self.UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED ,
			'UAC_TEMP_DUPLICATE_ACCOUNT' : self.UAC_TEMP_DUPLICATE_ACCOUNT ,
			'UAC_NORMAL_ACCOUNT' : self.UAC_NORMAL_ACCOUNT ,
			'UAC_INTERDOMAIN_TRUST_ACCOUNT' : self.UAC_INTERDOMAIN_TRUST_ACCOUNT ,
			'UAC_WORKSTATION_TRUST_ACCOUNT' : self.UAC_WORKSTATION_TRUST_ACCOUNT ,
			'UAC_SERVER_TRUST_ACCOUNT' : self.UAC_SERVER_TRUST_ACCOUNT ,
			'UAC_DONT_EXPIRE_PASSWD' : self.UAC_DONT_EXPIRE_PASSWD ,
			'UAC_MNS_LOGON_ACCOUNT' : self.UAC_MNS_LOGON_ACCOUNT ,
			'UAC_SMARTCARD_REQUIRED' : self.UAC_SMARTCARD_REQUIRED ,
			'UAC_TRUSTED_FOR_DELEGATION' : self.UAC_TRUSTED_FOR_DELEGATION ,
			'UAC_NOT_DELEGATED' : self.UAC_NOT_DELEGATED ,
			'UAC_USE_DES_KEY_ONLY' : self.UAC_USE_DES_KEY_ONLY ,
			'UAC_DONT_REQUIRE_PREAUTH' : self.UAC_DONT_REQUIRE_PREAUTH ,
			'UAC_PASSWORD_EXPIRED' : self.UAC_PASSWORD_EXPIRED ,
			'UAC_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION' : self.UAC_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION ,
		}

	@staticmethod
	def from_aduser(u):
		user = JackDawADUser()
		user.sn = lf(getattr(u,'sn'))
		user.cn = lf(getattr(u,'cn'))
		user.dn = lf(getattr(u,'distinguishedName'))
		user.initials = lf(getattr(u,'initials'))
		user.givenName = lf(getattr(u,'givenName'))
		user.displayName = lf(getattr(u,'displayName'))
		user.name = lf(getattr(u,'name'))
		user.objectCategory = lf(getattr(u,'objectCategory'))
		user.objectClass = lf(getattr(u,'objectClass'))
		user.objectGUID = lf(getattr(u,'objectGUID'))
		user.objectSid = lf(getattr(u,'objectSid'))
		user.primaryGroupID = lf(getattr(u,'primaryGroupID'))
		user.sAMAccountName = lf(getattr(u,'sAMAccountName'))
		user.userPrincipalName = lf(getattr(u,'userPrincipalName'))
		user.servicePrincipalName = lf(getattr(u,'servicePrincipalName'))
	
		user.memberOf = lf(getattr(u,'memberOf'))
		user.member = lf(getattr(u,'member'))
		user.accountExpires = dt(lf(getattr(u,'accountExpires')))
		user.badPasswordTime = dt(lf(getattr(u,'badPasswordTime')))
		user.lastLogoff = dt(lf(getattr(u,'lastLogoff')))
		user.lastLogon = dt(lf(getattr(u,'lastLogon')))
		user.lastLogonTimestamp = dt(lf(getattr(u,'lastLogonTimestamp')))
		user.pwdLastSet = dt(lf(getattr(u,'pwdLastSet')))
		user.whenChanged = dt(lf(getattr(u,'whenChanged')))
		user.whenCreated = dt(lf(getattr(u,'whenCreated')))
		user.badPwdCount = lf(getattr(u,'badPwdCount'))
		user.logonCount = lf(getattr(u,'logonCount'))
		user.sAMAccountType = lf(getattr(u,'sAMAccountType'))
		user.userAccountControl = lf(getattr(u,'userAccountControl'))
	
		user.codePage = lf(getattr(u,'codePage'))
		user.countryCode = lf(getattr(u,'countryCode'))
		user.when_pw_change = dt(lf(getattr(u,'when_pw_change')))
		user.when_pw_expires = dt(lf(getattr(u,'when_pw_expires')))
		user.must_change_pw = dt(lf(getattr(u,'must_change_pw')))
		user.canLogon = bc(lf(getattr(u,'canLogon')))
		user.isAdmin = bc(lf(getattr(u,'isAdmin', None)))
		calc_uac_flags(user)
			
		return user



from flask import current_app
from jackdaw.dbmodel.aduser import JackDawADUser

def list_users(domainid):
    db = current_app.db
    domains = []
    for uid, sid, name in db.session\
                            .query(JackDawADUser)\
                                .filter_by(ad_id = domainid)\
                                .with_entities(JackDawADUser.id, JackDawADUser.objectSid, JackDawADUser.sAMAccountName)\
                                .all():
        domains.append((uid, sid, name))
    return domains

def get(domainid, userid):
    db = current_app.db
    user = db.session.query(JackDawADUser).get(userid)
    return user.to_dict()

def get_sid(domainid, usersid):
    db = current_app.db
    for user in db.session.query(JackDawADUser).filter_by(objectSid = usersid).filter(JackDawADUser.ad_id == domainid).all():
        return user.to_dict()

def filter(domainid, proplist):
    #TODO: add other properties to search for!
    db = current_app.db
    query = db.session.query(JackDawADUser).filter_by(ad_id = domainid)
    for elem in proplist:
        if 'sAMAccountName' in elem:
            query = query.filter(JackDawADUser.sAMAccountName == elem['sAMAccountName'])
    
    user = query.first()
    if user is None:
        return {}
    return user.to_dict()


from . import Basemodel
import datetime
from sqlalchemy import Column, Integer, String, DateTime

class NetShare(Basemodel):
	__tablename__ = 'netshare'
	
	id = Column(Integer, primary_key=True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	machine_id = Column(Integer)
	ip = Column(String, index=True)
	rdns = Column(String, index=True)
	netname = Column(String, index=True)
	type = Column(String, index=True)
	remark = Column(String, index=True)
	passwd = Column(String, index=True)

	def to_dict(self):
		return {
			'id' : self.id , 
			'fetched_at' : self.fetched_at , 
			'machine_id' : self.machine_id , 
			'ip' : self.ip , 
			'rdns' : self.rdns , 
			'netname' : self.netname , 
			'type' : self.type , 
			'remark' : self.remark , 
			'passwd' : self.passwd ,
		}
	


from . import Basemodel, lf
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, ForeignKey, Index

class Credential(Basemodel):
	__tablename__ = 'credentials'
	__table_args__ = (Index('Credential_uc', "ad_id", "domain", "username","nt_hash", "lm_hash", "history_no", unique = True), )

	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer)
	domain = Column(String, index=True, nullable= False)
	username = Column(String, index=True, nullable= False)
	nt_hash = Column(String, index=True, nullable= False)
	lm_hash = Column(String, index=True, nullable= False)
	history_no = Column(Integer, index=True, nullable= False)
	
	def __init__(self, domain = None, username = None, nt_hash = None, lm_hash = None, history_no = None, ad_id = -1):
		self.domain = domain
		self.username = username
		self.nt_hash = nt_hash
		self.lm_hash = lm_hash
		self.history_no = history_no
		self.ad_id = ad_id

	@staticmethod
	def from_impacket_file(filename, ad_id = -1):
		"""
		Remember that this doesnt populate the foreign keys!!! You'll have to do it separately!
		important: historyno will start at 0. This means all history numbers in the file will be incremented by one
		"""
		with open(filename, 'r') as f:
			for line in f:
				cred = Credential()
				userdomainhist, flags, lm_hash, nt_hash, *t = line.split(':')
				#parsing history
				m = userdomainhist.find('_history')
				history_no = 0
				if m != -1:
					history_no = int(userdomainhist.split('_history')[1]) + 1
					userdomainhist = userdomainhist.split('_history')[0]
				m = userdomainhist.find('\\')
				domain = '<LOCAL>'
				username = userdomainhist
				if m != -1:
					domain = userdomainhist.split('\\')[0]
					username = userdomainhist.split('\\')[1]
				cred.ad_id = ad_id
				cred.nt_hash = nt_hash
				cred.lm_hash = lm_hash
				cred.history_no = history_no
				cred.username = username
				cred.domain = domain
				yield cred



from flask import current_app
from jackdaw.dbmodel.adcomp import JackDawADMachine

def list_machines(domainid):
    db = current_app.db
    domains = []
    for uid, sid, name in db.session\
                            .query(JackDawADMachine)\
                                .filter_by(ad_id = domainid)\
                                .with_entities(JackDawADMachine.id, JackDawADMachine.objectSid, JackDawADMachine.sAMAccountName)\
                                .all():
        domains.append((uid, sid, name))
    return domains

def get(domainid, machineid):
    db = current_app.db
    machine = db.session.query(JackDawADMachine).get(machineid)
    return machine.to_dict()

def get_sid(domainid, sid):
    db = current_app.db
    for machine in db.session.query(JackDawADMachine).filter_by(objectSid = sid).filter(JackDawADMachine.ad_id == domainid).all():
        return machine.to_dict()




from jackdaw.dbmodel.netshare import *
from jackdaw.dbmodel import *
from jackdaw import logger

from winrecon.cf.c_functions import NetShareEnum
from winrecon.file_utils import *

from msldap.core.msldap import *
from msldap.ldap_objects import *

from multiprocessing import Process, Queue
from threading import Thread
import threading
from dns import resolver, reversename
import ipaddress


class ShareEnumThread(Thread):
	def __init__(self, inQ, outQ):
		Thread.__init__(self)
		self.inQ = inQ
		self.outQ = outQ
		
	def run(self):
		while True:
			target = self.inQ.get()
			if not target:
				break
			try:
				for share in NetShareEnum(target, level=1):
					self.outQ.put((target, share))
			except Exception as e:
				logger.debug('ShareEnumerator error: %s' % str(e))
				continue
		
class ShareEnumProc(Process):
	def __init__(self, inQ, outQ, threadcnt):
		Process.__init__(self)
		self.inQ = inQ
		self.outQ = outQ
		self.threadcnt = threadcnt
		self.threads = []
		
	def run(self):
		for i in range(self.threadcnt):
			t = ShareEnumThread(self.inQ, self.outQ)
			t.daemon = True
			t.start()
			self.threads.append(t)			
		for t in self.threads:
			t.join()
		
class SMResProc(Process):
	def __init__(self, outQ, sql_con, dns_server = None):
		Process.__init__(self)
		self.outQ = outQ
		self.rdns_table = {}
		self.dns_table = {}
		self.session = None
		self.conn = sql_con
		self.dns_server = dns_server
		
	def setup(self):
		self.session = get_session(self.conn)
	
	def rdns_lookup(self, ip):
		if ip not in self.rdns_table:
			dns_resolver = resolver.Resolver()
			if self.dns_server:
				dns_resolver.nameservers = [self.dns_server]
			try:
				answer = str(dns_resolver.query(reversename.from_address(ip), "PTR")[0])
			except Exception as e:
				#print(e)
				answer = 'NA'
				pass
				
			self.rdns_table[ip] = answer
		return self.rdns_table[ip]
		
	def ip_lookup(self, target):
		if target not in self.dns_table:
			dns_resolver = resolver.Resolver()
			if self.dns_server:
				dns_resolver.nameservers = [self.dns_server]
			try:
				answers = dns_resolver.query(target, 'A')
				for rdata in answers:
					self.dns_table[target] = rdata.address
			except Exception as e:
				logger.debug('ShareEnumerator error: %s' % str(e))
				self.dns_table[target] = None
		return self.dns_table[target]
		
	def run(self):
		self.setup()
		while True:
			result = self.outQ.get()
			if not result:
				break
			
			target, share = result
			try:
				ip = str(ipaddress.ip_address(target))
			except Exception as e:
				#print(e)
				ip = None
			
			if ip is None:
				rdns = target
				ip = self.ip_lookup(target)
			else:
				rdns = self.rdns_lookup(ip)
			
			ns = NetShare()
			ns.ip = ip
			ns.rdns = rdns
			ns.netname = share.netname
			ns.type    = share.type
			ns.remark  = share.remark
			ns.passwd  = share.passwd
			self.session.add(ns)
			self.session.commit()
			#print('%s %s %s %s' % (target, ip, rdns, ns.netname))
			

class ShareEnumerator:
	def __init__(self, db_con, dns_server = None):
		self.db_con = db_con
		self.hosts = []
		self.inQ = Queue()
		self.outQ = Queue()
		self.agents = []
		self.result_process = None
		self.dns_server = dns_server
		
		self.agent_proccnt = 4
		self.agent_threadcnt = 4
		
		
	def load_targets_ldap(self, ldap):
		ldap_filter = r'(&(sAMAccountType=805306369))'

		attributes = ['sAMAccountName']
		for entry in ldap.pagedsearch(ldap_filter, attributes):
			self.hosts.append(entry['attributes']['sAMAccountName'][:-1])
			
	def load_targets_file(self, filename):
		with open(filename,'r') as f:
			for line in f:
				line=line.strip()
				if line == '':
					continue
				self.hosts.append(line)
				
	def load_tagets(self, targets):
		self.hosts = targets
		
	def run(self):
		create_db(self.db_con)

		self.result_process = SMResProc(self.outQ, self.db_con, dns_server = self.dns_server)
		self.result_process.daemon = True
		self.result_process.start()
		
		for i in range(self.agent_proccnt):
			p = ShareEnumProc(self.inQ, self.outQ, self.agent_threadcnt)
			p.daemon = True
			p.start()
			self.agents.append(p)
		
		logger.info('=== Enumerating shares ===')
		for t in self.hosts:
			self.inQ.put(t)
		
		for a in self.agents:
			for i in range(self.agent_threadcnt):
				self.inQ.put(None)
			
		for a in self.agents:
			a.join()
		
		self.outQ.put(None)
		self.result_process.join()
		



from jackdaw.dbmodel import *
from jackdaw import logger
from winrecon.cf.c_functions import NetSessionEnum
from winrecon.file_utils import *

from multiprocessing import Process, Queue
from threading import Thread
import threading
from dns import resolver, reversename
import time

class SessMonThread(Thread):
	def __init__(self, inQ, outQ):
		Thread.__init__(self)
		self.inQ = inQ
		self.outQ = outQ
		
	def run(self):
		while True:
			target = self.inQ.get()
			if not target:
				break
			try:
				for session in NetSessionEnum(target[:-1]):
					self.outQ.put((target, session))
			except Exception as e:
				logger.debug('SessionMonitor error: %s' % str(e))
				continue
		
class SessMonProc(Process):
	def __init__(self, inQ, outQ, threadcnt):
		Process.__init__(self)
		self.inQ = inQ
		self.outQ = outQ
		self.threadcnt = threadcnt
		self.threads = []
		
	def run(self):
		for i in range(self.threadcnt):
			t = SessMonThread(self.inQ, self.outQ)
			t.daemon = True
			t.start()
			self.threads.append(t)			
		for t in self.threads:
			t.join()
		
class SessMonResProc(Process):
	def __init__(self, outQ, sql_con, dns_server = None):
		Process.__init__(self)
		self.outQ = outQ
		self.rdns_table = {}
		self.session = None
		self.conn = sql_con
		self.dns_server = dns_server
		
	def setup(self):
		self.session = get_session(self.conn)
	
	def rdns_lookup(self, ip):
		if ip not in self.rdns_table:
			dns_resolver = resolver.Resolver()
			if self.dns_server:
				dns_resolver.nameservers = [self.dns_server]
			try:
				answer = str(dns_resolver.query(reversename.from_address(ip), "PTR")[0])
			except Exception as e:
				#print(e)
				answer = 'NA'
				pass
				
			self.rdns_table[ip] = answer
		return self.rdns_table[ip]
		
		
	def run(self):
		self.setup()
		while True:
			result = self.outQ.get()
			if not result:
				break
			
			target, session = result
			ip = session.computername.replace('\\\\','')
			
			ns = NetSession()
			ns.source = target
			ns.ip = ip
			ns.rdns = self.rdns_lookup(ip)
			ns.username = session.username
			self.session.add(ns)
			self.session.commit()
			#print('%s: %s\\%s (%s)' % (target, ip, session.username, ns.rdns))
		
class SessionMonitor:
	def __init__(self, db_conn, monitor_time = 60):
		self.db_conn = db_conn
		self.monitor_time = monitor_time
		self.hosts = []
		self.inQ = Queue()
		self.outQ = Queue()
		self.agents = []
		self.result_process = None
		
		self.agent_threadcnt = 4
		self.agent_proccnt = 4
		
	def load_targets_ldap(self, ldap):
		ldap_filter = r'(&(sAMAccountType=805306369))'

		attributes = ['sAMAccountName']
		for entry in ldap.pagedsearch(ldap_filter, attributes):
			self.hosts.append(entry['attributes']['sAMAccountName'])

	def load_targets_file(self, filename):
		with open(filename,'r') as f:
			for line in f:
				line=line.strip()
				if line == '':
					continue
				self.hosts.append(line)
				
	def load_tagets(self, targets):
		self.hosts = targets
			
	def run(self):
		create_db(self.db_conn)
		
		self.result_process = SessMonResProc(self.outQ, self.db_conn, dns_server = None)
		self.result_process.daemon = True
		self.result_process.start()
		
		for i in range(self.agent_proccnt):
			p = SessMonProc(self.inQ, self.outQ, self.agent_threadcnt)
			p.daemon = True
			p.start()
			self.agents.append(p)
		
		while True:
			#print('=== Polling sessions ===')
			logger.info('=== Polling sessions ===')
			for t in self.hosts:
				self.inQ.put(t)
			if self.monitor_time != -1:
				time.sleep(self.monitor_time)
				break
			
			time.sleep(10)
		
		for a in self.agents:
			for i in range(self.agent_threadcnt):
				self.inQ.put(None)
		
		
		for a in self.agents:
			a.join()
			
		self.outQ.put(None)
		self.result_process.join()
			
		
	


from . import Basemodel
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class JackDawTokenGroup(Basemodel):
	__tablename__ = 'tokengroup'
	
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="group_lookups", lazy = True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	cn = Column(String, index=True)
	dn = Column(String, index=True)
	guid = Column(String, index=True)
	sid = Column(String, index=True)
	member_sid = Column(String, index=True)
	is_group = Column(Boolean)
	is_user = Column(Boolean)
	is_machine = Column(Boolean)




import os
from jackdaw.dbmodel.spnservice import JackDawSPNService
from jackdaw.dbmodel.addacl import JackDawADDACL
from jackdaw.dbmodel.adgroup import JackDawADGroup
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adou import JackDawADOU
from jackdaw.dbmodel.usergroup import JackDawGroupUser
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.tokengroup import JackDawTokenGroup
from jackdaw.dbmodel import *
from jackdaw.wintypes.lookup_tables import *
from jackdaw import logger

#from msldap.core.common import *
from msldap.ldap_objects import *

class LDAPEnumerator:
	def __init__(self, db_conn, ldap):
		self.db_con = db_conn
		self.ldap = ldap
		
	def spnservice_enumerator(self):
		ldap_filter = r'(&(sAMAccountType=805306369))'
		attributes = ['sAMAccountName', 'servicePrincipalName']
		
		for entry in self.ldap.pagedsearch(ldap_filter, attributes):
			for spn in entry['attributes']['servicePrincipalName']:			
				port = None
				service, t = spn.rsplit('/',1)
				m = t.find(':')
				if m != -1:
					computername, port = spn.rsplit(':',1)
				else:
					computername = t
					
				s = JackDawSPNService()
				s.computername = computername
				s.service = service
				s.port = port
				yield s
			
	def get_domain_info(self):
		info = self.ldap.get_ad_info()
		return JackDawADInfo.from_dict(info.to_dict())
		
	def get_all_machines(self):
		for machine in self.ldap.get_all_machine_objects():
			yield (machine, JackDawADMachine.from_adcomp(machine))
			
	def get_all_users(self):
		for user in self.ldap.get_all_user_objects():
			#TODO: fix this ugly stuff here...
			if user.sAMAccountName[-1] == "$":
				continue
			yield (user, JackDawADUser.from_aduser(user))
			
			
	def get_all_ous(self):
		for ou in self.ldap.get_all_ous():
			yield (ou, JackDawADOU.from_adou(ou))

		
	def get_all_groups(self):
		for group in self.ldap.get_all_groups():
			yield JackDawADGroup.from_dict(group.to_dict())

		
	def get_user_effective_memberships(self, user):
		for sid in self.ldap.get_tokengroups(user.dn):
			s = JackDawTokenGroup()
			s.cn = str(user.cn)
			s.dn = str(user.dn)
			
			if isinstance(user, JackDawADUser):
				s.guid = str(user.objectGUID)
				s.sid = str(user.objectSid)
				s.member_sid = sid
				s.is_user = True
			elif isinstance(user, JackDawADMachine):
				s.guid = str(user.objectGUID)
				s.sid = str(user.objectSid)
				s.member_sid = sid
				s.is_machine = True
			elif isinstance(user, JackDawADGroup):
				s.guid = str(user.guid)
				s.sid = str(user.sid)
				s.member_sid = sid
				s.is_group = True		
				
			yield s
			
	def ace_to_dbo(self, obj, sd):
		if isinstance(obj, JackDawADUser):
			obj_type = 'user'
		elif isinstance(obj, JackDawADMachine):
			obj_type = 'machine'
		elif isinstance(obj, JackDawADGroup):
			obj_type = 'group'
		elif isinstance(obj, JackDawADOU):
			obj_type = 'ou'
		else:
			raise Exception('Unknown object type %s' % type(obj))
		
		order_ctr = 0
		for ace in sd.nTSecurityDescriptor.Dacl.aces:
			acl = JackDawADDACL()
			acl.object_type = obj_type
			acl.object_type_guid = OBJECTTYPE_GUID_MAP.get(obj_type)
			acl.owner_sid = str(sd.nTSecurityDescriptor.Owner)
			acl.group_sid = str(sd.nTSecurityDescriptor.Group)
			acl.ace_order = order_ctr
			
			order_ctr += 1
			acl.guid = str(sd.objectGUID)
			if sd.objectSid:
				acl.sid = str(sd.objectSid)
			if sd.cn:
				acl.cn = sd.cn
			if sd.distinguishedName:
				acl.dn = str(sd.distinguishedName)
			acl.sd_control = sd.nTSecurityDescriptor.Control
			
			acl.ace_type = ace.Header.AceType.name
			acl.ace_mask = ace.Mask
			t = getattr(ace,'ObjectType', None)
			if t:
				acl.ace_objecttype = str(t)
			
			t = getattr(ace,'InheritedObjectType', None)
			if t:
				acl.ace_inheritedobjecttype = str(t)
				
			true_attr, false_attr = JackDawADDACL.mask2attr(ace.Mask)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
				
			true_attr, false_attr = JackDawADDACL.hdrflag2attr(ace.Header.AceFlags)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
			
			acl.ace_sid = str(ace.Sid)
			yield acl
				
	def get_acls_for_dn(self, obj):
		for sd in self.ldap.get_objectacl_by_dn(obj.dn):
			if not sd.nTSecurityDescriptor or not sd.nTSecurityDescriptor.Dacl:
				continue
			
			for acl in self.ace_to_dbo(obj, sd):
				yield acl
		
	def get_current_dc_ip(self):
		pass
		#print(os.environ['LOGONSERVER'])
	
	@staticmethod
	def spn_to_account(spn):
		if spn.find('/') != -1:
			return spn.rsplit('/')[1].upper() + '$'
		
	def run(self):
		self.ldap.connect()
		session = get_session(self.db_con)
		info = self.get_domain_info()
		session.add(info)
		session.commit()
		session.refresh(info)
		
		logger.info('Enumerating users...')
		for obj, user in self.get_all_users():
			user.ad_id = info.id
			session.add(user)
			session.commit()
			session.refresh(user)
			
			for membership in self.get_user_effective_memberships(user):
				info.group_lookups.append(membership)
				
			for acl in self.get_acls_for_dn(user):
				acl.ad_id = info.id
				session.add(acl)
			
			for spn in getattr(obj,'allowedtodelegateto',[]):
				con = JackDawUserConstrainedDelegation()
				con.spn = spn
				con.targetaccount = self.spn_to_account(spn)
				user.allowedtodelegateto.append(con)
			
			session.commit()
		
		logger.info('Enumerating machines...')
		for obj, machine in self.get_all_machines():
			machine.ad_id = info.id
			session.add(machine)
			session.commit()
			session.refresh(machine)
			
			for membership in self.get_user_effective_memberships(machine):
				info.group_lookups.append(membership)
			
			for spn in getattr(obj,'allowedtodelegateto',[]):
				con = JackDawMachineConstrainedDelegation()
				con.spn = spn
				con.targetaccount = self.spn_to_account(spn)
				machine.allowedtodelegateto.append(con)
			
			for acl in self.get_acls_for_dn(machine):
				acl.ad_id = info.id
				session.add(acl)
				
			session.commit()
			
		logger.info('Enumerating groups...')
		ctr = 0		
		for group in self.get_all_groups():
			group.ad_id = info.id	
			session.add(group)
			
			for membership in self.get_user_effective_memberships(group):
				info.group_lookups.append(membership)
				
			for acl in self.get_acls_for_dn(group):
				acl.ad_id = info.id
				session.add(acl)
				
			if ctr % 1000 == 0:
				session.commit()
				
		logger.info('Enumerating OUs...')
		for obj, ou in self.get_all_ous():
			ou.ad_id = info.id
			session.add(ou)
			session.commit()
			session.refresh(ou)
				
			for acl in self.get_acls_for_dn(ou):
				acl.ad_id = info.id
				session.add(acl)
				
			session.commit()
			
		logger.info('Enumerating service users')
		for spnservice in self.spnservice_enumerator():
			spnservice.ad_id = info.id	
			session.add(spnservice)
		session.commit()
		
		"""
		ctr = 0		
		for acl in self.get_all_acls():
			acl.ad_id = info.id
			session.add(acl)
			if ctr % 1000 == 0:
				session.commit()
		"""
		session.commit()
		
		return info.id
		
		


from . import Basemodel, lf
import datetime
from sqlalchemy import Column, Integer, String

class JackDawGroupUser(Basemodel):
	__tablename__ = 'groupuser'

	id = Column(Integer, primary_key=True)
	group_osid = Column(String, index=True)
	user_osid = Column(String, index=True)
	
	
	def __init__(self, group_osid, user_osid):
		self.group_osid = group_osid
		self.user_osid = user_osid
		


from flask import current_app
from jackdaw.dbmodel.adou import JackDawADOU

def list_ous(domainid):
    db = current_app.db
    domains = []
    for uid, guid, ou in db.session\
                            .query(JackDawADOU)\
                                .filter_by(ad_id = domainid)\
                                .with_entities(JackDawADOU.id, JackDawADOU.objectGUID, JackDawADOU.ou)\
                                .all():
        domains.append((uid, guid, ou))
    return domains

def get(domainid, ouid):
    db = current_app.db
    user = db.session.query(JackDawADOU).get(ouid)
    return user.to_dict()

def get_guid(domainid, guid):
    db = current_app.db
    for user in db.session.query(JackDawADOU).filter_by(objectGUID = guid).filter(JackDawADOU.ad_id == domainid).all():
        return user.to_dict()


from setuptools import setup, find_packages

setup(
	# Application name:
	name="jackdaw",

	# Version number (initial):
	version="0.1.1",

	# Application author details:
	author="Tamas Jos",
	author_email="info@skelsec.com",

	# Packages
	packages=find_packages(),

	# Include additional files into the package
	include_package_data=True,
	
	# Details
	url="https://github.com/skelsec/jackdaw",

	zip_safe = False,
	#
	# license="LICENSE.txt",
	description="Gathering shiny things from your domain",

	# long_description=open("README.txt").read(),
	python_requires='>=3.6',
	classifiers=(
		"Programming Language :: Python :: 3.6",
		"Operating System :: OS Independent",
	),
	install_requires=[
		'aiosmb>=0.1.5',
		'msldap>=0.2.3',
		'sqlalchemy',
		'dnspython',
		'tqdm',
		'networkx',
		'connexion',
		'flask-sqlalchemy',
		'connexion[swagger-ui]'
	],
	entry_points={
		'console_scripts': [
			'jackdaw = jackdaw.__main__:main',
		],
	}
)



import json
import enum
import ipaddress
import datetime
from connexion.apps.flask_app import FlaskJSONEncoder

class UniversalFlaskEncoder(FlaskJSONEncoder):
	"""
	Used to override the default json encoder to provide a direct serialization for formats
	that the default json encoder is incapable to serialize
	"""
	def default(self, obj):
		if isinstance(obj, datetime.datetime):
			return obj.isoformat()
		elif isinstance(obj, enum.Enum):
			return obj.value
		elif isinstance(obj, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
			return str(obj)
		elif hasattr(obj, 'to_dict'):
			return obj.to_dict()
		else:
			return json.JSONEncoder.default(self, obj)


from . import Basemodel, lf, dt
from jackdaw.dbmodel.utils import *
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class JackDawADMachine(Basemodel):
	__tablename__ = 'machines'

	# Now for the attributes
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="computers", lazy = True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	sn = Column(String)
	cn = Column(String)
	dn = Column(String)
	accountExpires = Column(DateTime)
	badPasswordTime = Column(DateTime)
	badPwdCount = Column(String)
	codePage = Column(String)
	countryCode = Column(String)
	displayName = Column(String)
	dNSHostName = Column(String)
	instanceType = Column(String)
	isCriticalSystemObject = Column(String)
	lastLogoff =Column(DateTime)
	lastLogon = Column(DateTime)
	lastLogonTimestamp = Column(DateTime)
	logonCount = Column(Integer)
	localPolicyFlags = Column(String)
	supported_enc_types = Column(Integer)
	name = Column(String)
	nTSecurityDescriptor = Column(String)
	objectCategory = Column(String)
	objectClass = Column(String)
	objectGUID = Column(String)
	objectSid = Column(String)
	operatingSystem = Column(String)
	operatingSystemVersion = Column(String)
	primaryGroupID = Column(String)
	pwdLastSet = Column(DateTime)
	sAMAccountName = Column(String)
	sAMAccountType = Column(String)
	userAccountControl = Column(Integer)
	whenChanged = Column(DateTime)
	whenCreated = Column(DateTime)
	servicePrincipalName = Column(String)
	
	allowedtodelegateto = relationship("JackDawMachineConstrainedDelegation", back_populates="machine", lazy = True)
	
	UAC_SCRIPT = Column(Boolean)
	UAC_ACCOUNTDISABLE = Column(Boolean)
	UAC_HOMEDIR_REQUIRED = Column(Boolean)
	UAC_LOCKOUT = Column(Boolean)
	UAC_PASSWD_NOTREQD = Column(Boolean)
	UAC_PASSWD_CANT_CHANGE = Column(Boolean)
	UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED = Column(Boolean)
	UAC_TEMP_DUPLICATE_ACCOUNT = Column(Boolean)
	UAC_NORMAL_ACCOUNT = Column(Boolean)
	UAC_INTERDOMAIN_TRUST_ACCOUNT = Column(Boolean)
	UAC_WORKSTATION_TRUST_ACCOUNT = Column(Boolean)
	UAC_SERVER_TRUST_ACCOUNT = Column(Boolean)
	UAC_NA_1 = Column(Boolean)
	UAC_NA_2 = Column(Boolean)
	UAC_DONT_EXPIRE_PASSWD = Column(Boolean)
	UAC_MNS_LOGON_ACCOUNT = Column(Boolean)
	UAC_SMARTCARD_REQUIRED = Column(Boolean)
	UAC_TRUSTED_FOR_DELEGATION = Column(Boolean)
	UAC_NOT_DELEGATED = Column(Boolean)
	UAC_USE_DES_KEY_ONLY = Column(Boolean)
	UAC_DONT_REQUIRE_PREAUTH = Column(Boolean)
	UAC_PASSWORD_EXPIRED = Column(Boolean)
	UAC_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION = Column(Boolean)


	def to_dict(self):
		return {
			'id' : self.id ,
			'ad_id' : self.ad_id ,
			'dn' : self.dn ,
			'displayName' : self.displayName ,
			'name' : self.name ,
			'objectSid' : self.objectSid ,
			'sAMAccountName' : self.sAMAccountName ,
			'servicePrincipalName' : self.servicePrincipalName ,
			'accountExpires' : self.accountExpires ,
			'badPasswordTime' : self.badPasswordTime ,
			'lastLogoff' : self.lastLogoff ,
			'lastLogon' : self.lastLogon ,
			'lastLogonTimestamp' : self.lastLogonTimestamp ,
			'pwdLastSet' : self.pwdLastSet ,
			'whenChanged' : self.whenChanged ,
			'whenCreated' : self.whenCreated ,
			'badPwdCount' : self.badPwdCount ,
			'logonCount' : self.logonCount ,
			'sAMAccountType' : self.sAMAccountType ,
			'userAccountControl' : self.userAccountControl ,
			'codePage' : self.codePage ,
			'countryCode' : self.countryCode ,
			'supported_enc_types' : self.supported_enc_types ,
			'localPolicyFlags' : self.localPolicyFlags ,
			'operatingSystem' : self.operatingSystem ,
			'operatingSystemVersion' : self.operatingSystemVersion ,
			'primaryGroupID' : self.primaryGroupID ,
			'UAC_SCRIPT' : self.UAC_SCRIPT ,
			'UAC_ACCOUNTDISABLE' : self.UAC_ACCOUNTDISABLE ,
			'UAC_HOMEDIR_REQUIRED' : self.UAC_HOMEDIR_REQUIRED ,
			'UAC_LOCKOUT' : self.UAC_LOCKOUT ,
			'UAC_PASSWD_NOTREQD' : self.UAC_PASSWD_NOTREQD ,
			'UAC_PASSWD_CANT_CHANGE' : self.UAC_PASSWD_CANT_CHANGE ,
			'UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED' : self.UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED ,
			'UAC_TEMP_DUPLICATE_ACCOUNT' : self.UAC_TEMP_DUPLICATE_ACCOUNT ,
			'UAC_NORMAL_ACCOUNT' : self.UAC_NORMAL_ACCOUNT ,
			'UAC_INTERDOMAIN_TRUST_ACCOUNT' : self.UAC_INTERDOMAIN_TRUST_ACCOUNT ,
			'UAC_WORKSTATION_TRUST_ACCOUNT' : self.UAC_WORKSTATION_TRUST_ACCOUNT ,
			'UAC_SERVER_TRUST_ACCOUNT' : self.UAC_SERVER_TRUST_ACCOUNT ,
			'UAC_DONT_EXPIRE_PASSWD' : self.UAC_DONT_EXPIRE_PASSWD ,
			'UAC_MNS_LOGON_ACCOUNT' : self.UAC_MNS_LOGON_ACCOUNT ,
			'UAC_SMARTCARD_REQUIRED' : self.UAC_SMARTCARD_REQUIRED ,
			'UAC_TRUSTED_FOR_DELEGATION' : self.UAC_TRUSTED_FOR_DELEGATION ,
			'UAC_NOT_DELEGATED' : self.UAC_NOT_DELEGATED ,
			'UAC_USE_DES_KEY_ONLY' : self.UAC_USE_DES_KEY_ONLY ,
			'UAC_DONT_REQUIRE_PREAUTH' : self.UAC_DONT_REQUIRE_PREAUTH ,
			'UAC_PASSWORD_EXPIRED' : self.UAC_PASSWORD_EXPIRED ,
			'UAC_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION' : self.UAC_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION ,
		}
	
	@staticmethod
	def from_adcomp(u):
		machine = JackDawADMachine()
		machine.sn = lf(getattr(u,'sn'))
		machine.cn = lf(getattr(u,'cn'))
		machine.dn = lf(getattr(u,'distinguishedName'))
		machine.accountExpires = dt(lf(getattr(u,'accountExpires')))
		machine.badPasswordTime = dt(lf(getattr(u,'badPasswordTime')))
		machine.badPwdCount = lf(getattr(u,'badPwdCount'))
		machine.codePage = lf(getattr(u,'codePage'))
		machine.countryCode = lf(getattr(u,'countryCode'))
		machine.displayName = lf(getattr(u,'displayName'))
		machine.dNSHostName = lf(getattr(u,'dNSHostName'))
		machine.instanceType = lf(getattr(u,'instanceType'))
		machine.isCriticalSystemObject = lf(getattr(u,'isCriticalSystemObject'))
		machine.lastLogoff = dt(lf(getattr(u,'lastLogoff')))
		machine.lastLogon = dt(lf(getattr(u,'lastLogon')))
		machine.lastLogonTimestamp = dt(lf(getattr(u,'lastLogonTimestamp')))
		machine.logonCount = lf(getattr(u,'logonCount'))
		machine.localPolicyFlags = lf(getattr(u,'localPolicyFlags'))
		machine.supported_enc_types = lf(getattr(u,'supported_enc_types'))
		machine.name = lf(getattr(u,'name'))
		machine.nTSecurityDescriptor = lf(getattr(u,'nTSecurityDescriptor'))
		machine.objectCategory = lf(getattr(u,'objectCategory'))
		machine.objectClass = lf(getattr(u,'objectClass'))
		machine.objectGUID = lf(getattr(u,'objectGUID'))
		machine.objectSid = lf(getattr(u,'objectSid'))
		machine.operatingSystem = lf(getattr(u,'operatingSystem'))
		machine.operatingSystemVersion = lf(getattr(u,'operatingSystemVersion'))
		machine.primaryGroupID = lf(getattr(u,'primaryGroupID'))
		machine.pwdLastSet = dt(lf(getattr(u,'pwdLastSet')))
		machine.sAMAccountName = lf(getattr(u,'sAMAccountName'))
		machine.sAMAccountType = lf(getattr(u,'sAMAccountType'))
		machine.userAccountControl = lf(int(getattr(u,'userAccountControl', 0)))
		machine.whenChanged = dt(lf(getattr(u,'whenChanged')))
		machine.whenCreated = dt(lf(getattr(u,'whenCreated')))
		machine.servicePrincipalName = lf(getattr(u,'servicePrincipalName'))
		calc_uac_flags(machine)
		
		return machine